/*
 * ADsymbolic, created by Henry Chang  -  Fall 1990
 * Last modified:  March 1992  -  Enrico Malavasi */

#define C_FILE_NAME "ADsymbolic.c"

#include <stdio.h>
#include "ADglobals.h"
#include "oct.h"
#include "errtrap.h"
#include "symlib.h"
#include "oh.h"
#include "nle.h"
#include "tap.h"
#include "vulcan.h"

#include "TEKtech.h"			/* need it for technName        */

#define Y_OFFSET 2000			/* 100 lambda offset            */ 
#define MAX_OCT_NAME 100		/* Maximum chars in oct name    */
#define CONNECTOR "com2.4x4"		/* Hard wired connector for now */
					/* It must be in the leaf cells */
					/* directory.			*/
#define CONNECTOR_OFFSET 600		/* Offset to max_x, 30 lambdas  */
#define SNAP_OCT_VALUE 20		/* Round to nearest 20 oct      */
                                        /* in placement                 */

#define MOSAICO_HELP 1
#define MOSAICO_OFFSET 200

#define POWER_NET -1

extern octObject *makebox();

typedef struct _ADSnetlist {
    octObject 	      net;
    int		      number;		/* Number assigned to net */
    struct _ADSnetlist *next;
} _ADSnetlist;

static void
  findRealBB( mod, BB )
struct _module *mod;
octBox         *BB;
{
    octObject  facet;

    ohUnpackDefaults (&facet,"r",":physical:contents");
    ohUnpackFacetName (&facet, mod->name);
    OH_ASSERT( octOpenFacet( &facet ));
    ohBB( &facet, BB );
    octCloseFacet( &facet );
}

ADSsquareConn( facet, instance, layerName, width, height, instLib )
octObject *facet;
octObject *instance;
char      *layerName;
octCoord   width;
octCoord   height;
char      *instLib;
{
    char name[256];
    octObject term;
    octObject master;

    sprintf(name, "%s/%s.%dx%d",
	    instLib, layerName, width/coordSize, height/coordSize);
    ohUnpackDefaults (&master, "w", ":physical:contents");
    ohUnpackFacetName (&master, name);
    symOpenFacet( &master, "w", technName );

    ohCreateTerm( &master, &term, "t" );
    octAttach( &term, makebox(&master, layerName, -width/2, -height/2, width/2, height/2));

    octCloseFacet( &master );
    ADSgenInterface (&master);

    ohCreateInstance(facet, instance, "", name, "physical", 0, 0, OCT_NO_TRANSFORM);
}

ADSgenInterface (facet)
octObject *facet;
{
    octObject iface;
    vulcanMessageType message;

    iface.type = OCT_FACET;
    iface.contents.facet.cell = facet->contents.facet.cell;
    iface.contents.facet.view = facet->contents.facet.view;
    iface.contents.facet.facet = "interface";
    iface.contents.facet.version = facet->contents.facet.version;

    VERBOSE (printf("Creating Interface facet: %s:%s:%s\n",
	iface.contents.facet.cell,
        iface.contents.facet.view,
        iface.contents.facet.facet));

    iface.contents.facet.mode = "w";
    facet->contents.facet.mode = "r";

    OH_ASSERT(octOpenFacet(facet));
    OH_ASSERT(octOpenFacet(&iface));

    vulcanMessageHandler (NULL);
    message = vulcanParameterized (facet,&iface,VULCAN_DESTROY_OLD,
				   VULCAN_DFLT_MIN_FEATURE, VULCAN_NO_DEBUG);

    OH_ASSERT(octCloseFacet(facet));
    OH_ASSERT(octCloseFacet(&iface));

}

ADScreateADSnetlist (facet,ADSnetlist)
octObject *facet;
_ADSnetlist **ADSnetlist;
{
    struct _net *net;
    int 	counter;
    _ADSnetlist *ADSnet, *previous;
    char netname[MAX_OCT_NAME];
    
    *ADSnetlist = NULL;
    counter = 0;
    for (net=netlist; net!=NULL; net=net->next) {
	if (net->firstnett->nxtnett != NULL) {
	    CALLOC_ERRCHK (ADSnet, 1, _ADSnetlist);
	    if (counter==0) {
		*ADSnetlist = ADSnet;
	    } /* first one */ 
	    else {
		previous->next = ADSnet; 
	    } /* Following ones */ 
	    ADSnet->next = NULL;
	    ADSnet->number = net->number;
	    if(net->name[0] == '\0') {
		sprintf (netname, "net_%d", net->number);
		sprintf(stderr, "*** WARNING: Input net with no name renamed %s",
			netname);
	    } else
	      strcpy(netname, net->name);
	    OH_ASSERT( ohCreateNet (facet, &(ADSnet->net), netname ));
	    previous = ADSnet;
	    counter++;
	} /* check to make sure that there are at least two terminals 
	     on the net */ 
    } /* for: net search */
} /* ADScreateADSnetlist */

ADScheckADSnetlist (ADSnetlist)
_ADSnetlist *ADSnetlist;
{
   _ADSnetlist *ADSnet;
   printf ("ADScheckADSnetlist:");
   for (ADSnet = ADSnetlist; ADSnet != NULL; ADSnet=ADSnet->next) {
      printf (" %d",ADSnet->number);
   }
   printf ("\n");
}

_ADSnetlist *ADSsearchADSnetlist (ADSnetlist,netnumber)
_ADSnetlist *ADSnetlist;
int netnumber;
{
    _ADSnetlist *ADSnet;
    for (ADSnet = ADSnetlist; ADSnet != NULL; ADSnet = ADSnet -> next) {
       if (netnumber==ADSnet->number) {
          return (ADSnet); 
       }
    } /* for */
    return NULL;
} /* ADSsearchADSnetlist */

static void
  ADScreateCBLKifaces ()
/* Create Interface Facets for the CBLKs  */
{
   struct _module 	*mod;
   octObject		int_facet;

   for(mod=modlist[CBLK]; mod !=NULL; mod=mod->next) {
      ohUnpackDefaults (&int_facet,"r",":physical:contents");
      ohUnpackFacetName (&int_facet, mod->name);
      ADSgenInterface (&int_facet);
   }
}

ADSaddConstraint (sidebag,instance)
octObject *sidebag;
octObject *instance;
{
    octObject prop;
    octObject con_bag;
    OH_ASSERT(ohCreateBag(sidebag,&con_bag,"CONSTRAINT"));
    OH_ASSERT(octAttach (&con_bag,instance));
    OH_ASSERT(ohCreatePropInt(&con_bag,&prop,"MAX",0));
    OH_ASSERT(ohCreatePropInt(&con_bag,&prop,"MIN",0));

} /* ADSaddConstraint */

ADSaddConstraintBags (facet, topbag, bottombag, rightbag, leftbag) 
octObject *facet;
octObject *topbag;
octObject *bottombag;
octObject *rightbag;
octObject *leftbag;
{
    octObject xybag; 
    
    OH_ASSERT(ohCreateBag(facet,&xybag,"YCONSTRAINTS"));
    OH_ASSERT(ohCreateBag(&xybag,topbag,"TOP"));
    OH_ASSERT(ohCreateBag(&xybag,bottombag,"BOTTOM"));
    OH_ASSERT(ohCreateBag(facet,&xybag,"XCONSTRAINTS"));
    OH_ASSERT(ohCreateBag(&xybag,rightbag,"RIGHT"));
    OH_ASSERT(ohCreateBag(&xybag,leftbag,"LEFT"));
}

static void
  ADSgenClockWires( facet, instanceDir )
octObject *facet;
char      *instanceDir;
/* Generate the clock lines   -  Enrico Malavasi, March 1992 */
{
    struct _net *net;
    octObject   Net, Term;
    octBox      BB;
    struct {
	octObject term;
	octCoord   pos;
    } termList[200], connector;
    octCoord    y0, y, x;
    octObject   row, z;
    octObject   connBag;
    octObject   hPath, vPath, new;
    octObject   metalLayer, polyLayer;
    tapLayerListElement layerList[2];
    octGenerator gener;
    int  termListNum;
    int  i, j;

    OH_ASSERT( ohGetByInstName( facet, &row, "SWCH" ) );
    ohGetOrCreateBag( facet, &connBag, "CONNECTORS");
    ohGetOrCreateLayer( facet, &metalLayer, metal_L );
    ohGetOrCreateLayer( facet, &polyLayer,  botpoly_L );

    /* For each clock net, generate the list of upper terminals
     * in the switch row */
    for(net=netlist; net!=NULL; net=net->next) {
	if(net->type == CLOCK) {
	    termListNum = 0;
	    OH_ASSERT( ohGetByNetName( facet, &Net, net->name));
	    octInitGenContents( &Net, OCT_TERM_MASK, &gener );
	    while( octGenerate( &gener, &Term ) == OCT_OK ) {
		if(Term.contents.term.instanceId == 0) {
		    octGenFirstContent( &Term, OCT_TERM_MASK, &new);
		    Term = new;
		}
		ohBB( &Term, &BB );
		if(octIsAttached( &row, &Term ) == OCT_OK ) {		    
		    termList[termListNum].term = Term;
		    termList[termListNum].pos  = (BB.lowerLeft.x+BB.upperRight.x)/2;
		    y0 = (BB.lowerLeft.y+BB.upperRight.y)/2;
		    termListNum++;
		} else {
		    octGenFirstContainer( &Term, OCT_INSTANCE_MASK, &new);
		    connector.term = Term;
		    connector.pos  = (BB.lowerLeft.x+BB.upperRight.x)/2;
		    y = (BB.lowerLeft.y+BB.upperRight.y)/2;
		}
	    }
	    octFreeGenerator( &gener );

	    /* Sort the selected terminals */
	    for(i=0 ; i<termListNum ; i++) {
		for(j=i+1 ; j<termListNum ; j++) {
		    if(termList[i].pos > termList[j].pos) {
			x = termList[i].pos;
			z = termList[i].term;
			termList[i].pos = termList[j].pos;
			termList[i].term = termList[j].term;
			termList[j].pos = x;
			termList[j].term = z;
		    }
		}
	    }

	    /* Generate the clock lines */
	    for(i=0 ; i<termListNum ; i++) {
		/* Create the new via-connector */
		layerList[0].layer = metalLayer;
		layerList[0].width = metwidth;
		layerList[0].direction = TAP_HORIZONTAL;
		layerList[1].layer = polyLayer;
		layerList[1].width = polywidth;
		layerList[1].direction = TAP_VERTICAL;

		tapGetConnector( 2, layerList, &new );
		ohCreateInstance( facet, &new, "t",
				 new.contents.instance.master,
				 new.contents.instance.view,
				 termList[i].pos, y, OCT_NO_TRANSFORM);
		octAttachOnce( &connBag, &new );
		OH_ASSERT( octGenFirstContent(&new, OCT_TERM_MASK, &Term ) );

		/* Horizontal segment in metal */
		ohCreateSegment( &metalLayer, &hPath, connector.pos, y,
				termList[i].pos, y, metwidth);
		octAttach( &Net, &hPath );
		octAttach( &hPath, &(connector.term));
		octAttach( &hPath, &Term );

		/* Vertical segment in poly */
		ohCreateSegment( &polyLayer, &vPath, termList[i].pos, y0,
				termList[i].pos, y, polywidth);
		octAttach( &Net, &vPath );
		octAttach( &vPath, &(termList[i].term) );
		octAttach( &vPath, &Term );

		connector.pos  = termList[i].pos;
		connector.term = Term;
	    }
	}
    }
}

static void
  ADScreateRow (facet, first_module, name, y_position, final_x_position, ADSnetlist, instances, libDir)
/* Creates a new cell with a row of Opamps or Switches.
 * Then Instantiates that into facet. */
octObject 	*facet;
struct _module 	*first_module;
char      	*name;
int       	*y_position;
int       	*final_x_position;	
_ADSnetlist 	*ADSnetlist;
char	  	*instances;
char            *libDir;
{
    typedef struct _terms {
        char		terminalname[TERMINALNAME_SIZE];
        int		netnumber;		/* non-power nets          */
        char		netname[NETNAME_SIZE];  /* power nets,             */
                                                /* netnumber = POWER_NET   */
        char		property[PROPERTYNAME_SIZE];
                                                /* SUPPLY or GROUND        */
                                                /* netnumber = POWER_NET   */
        int		side;			/* Side of Power term      */
                                                /* netnumber = POWER_NET   */
        struct _terms 	*next;
    } _terms;

    typedef struct _snet {			/* Horizontal Power Net    */
        struct _terms   *term;
        octObject	net;
        int		numterms;		/* Number of terminals     */
        struct _snet    *next;
    } _snet;

    octObject	row;			        /* row facet to be created */
    char	cellname[FILENAME_SIZE];	/* row name                */
    char	*instname;			/* OCT name for instances  */

    octObject		int_facet; 		/* Intermediate facet      */
						/* Used for unpacking name */
    octObject		inst;			/* instance                */
    octBox              BB;
    octTransformType	transform;

    struct _module 	*mod;
    int			counter;
    int			x_position, next_x_position;
    int			max_y;
    char		modulename[FILENAME_SIZE];

    char		terminalname[TERMINALNAME_SIZE];
    struct _terminal	*term;			/* Signal Terminal         */
    struct _sterminal	*sterm;			/* Supply Terminal         */
    octObject		octterm,formalterm;
    struct _terms	*firstterm, *previousterm, *presentterm; 
    struct _snet	*firstnet, *previousnet, *presentnet, *searchnet;
    _ADSnetlist		*ADSnet;

    int			nummods;		/* Number of modules       */
    int			found;			/* Flag			   */
    octObject		prop;			/* Property		   */

    /*
     * Count the number of modules
     */
    nummods = 0;
    for (mod=first_module; mod != NULL; mod=mod->next) {
       nummods++;
    }

    MALLOC_ERRCHK (instname, char, strlen (name) + 5);
    /* 5 is arbitrary: allows for 4 char in the number, 1 is for `\0' */ 
  
    sprintf (cellname,"%s/%s",instances, name);
    ohUnpackDefaults (&row, "w", ":symbolic:contents");
    ohUnpackFacetName (&row,cellname);
    symOpenFacet (&row, "w", technName);

    firstterm = NULL;
    counter = 0;
    next_x_position = 0;
    max_y = 0;
    for (mod=first_module; mod != NULL; mod=mod->next) {

       /* Detect the real size of every element  */
       findRealBB( mod, &BB );
       mod->xdim = BB.upperRight.x - BB.lowerLeft.x;
       mod->ydim = BB.upperRight.y - BB.lowerLeft.y;

       counter++;				/* Start at 1 */
       mod->ytrans = *y_position;
       x_position = next_x_position;
       next_x_position += mod->xdim;
       if (mod->transform==NO) {
          transform = OCT_NO_TRANSFORM;
       } /* if */ 
       else if (mod->transform==YES) { 
          transform = OCT_MIRROR_X; 
          x_position += mod->xdim;
       } /* else if */
       else {
 	  errRaise (C_FILE_NAME,0,"mod->transform not YES or NO value.");
       } /* else */
 
       if ( mod->ydim > max_y ) {
          max_y = mod->ydim;
       } /* if */

       ohUnpackDefaults (&int_facet,"r",":physical:contents");
       sprintf(modulename, "../%s", mod->name);
       /*
        * This is what restricts the instances directory to be there
        * and to be in the present working directory.
        */
       ohUnpackFacetName (&int_facet, modulename);
       sprintf(instname, "%s.%d", name, counter);  
       inst.type = OCT_INSTANCE;
       inst.contents.instance.master = int_facet.contents.facet.cell;
       inst.contents.instance.view = int_facet.contents.facet.view;
       inst.contents.instance.facet = int_facet.contents.facet.facet;
       inst.contents.instance.version = int_facet.contents.facet.version;
       inst.contents.instance.transform.transformType = transform;
       inst.contents.instance.transform.translation.x = x_position;
       inst.contents.instance.transform.translation.y = 0;

       symCreateInstance (&row,&inst);
       inst.contents.instance.name = instname;
       octModify(&inst);

       /*
        * Promote terminals on instance to formal terminals
        */

       /*
        * Signal Terminals
        */

       for (term=mod->terminals; term!= NULL; term=term->next) {
           if (strcmp(term->name,"")!=0) {
               OH_ASSERT (ohGetByTermName (&inst, &octterm, term->name) );
               sprintf (terminalname,"%s.%d",term->name,counter);
               symCreateFormalTerm (&row,&formalterm,terminalname);
               symImplementFormalTerm (&formalterm,&octterm);
               previousterm = presentterm;
               CALLOC_ERRCHK(presentterm, 1, struct _terms); 
               if (firstterm==NULL) {
                  firstterm = presentterm;
               } /* if */
               else {
                  previousterm->next = presentterm; 
               } /* else */
               strcpy (presentterm->terminalname,terminalname);
               presentterm->netnumber = term->netptr->number; 
               strcpy (presentterm->netname,"");
               strcpy (presentterm->property,"");
               presentterm->side = 0;
               presentterm->next = NULL;
           } /* if */
       } /* for */ 

       /*
        * Supply Terminals
        */
       for (sterm=mod->sterminals; sterm!= NULL; sterm=sterm->next) {
           if ( (strcmp(sterm->name,"")!=0) &&
                ( (sterm->side == TOP) || (sterm->side == BOTTOM) ||
                  ( (transform == OCT_NO_TRANSFORM) && (sterm->side == LEFT) &&
                    (counter == 1) ) ||
                  ( (transform == OCT_NO_TRANSFORM) && (sterm->side==RIGHT) &&
                    (counter == nummods) ) ||
                  ( (transform == OCT_MIRROR_X) && (sterm->side == RIGHT) && 
                     (counter == 1) ) ||
                  ( (transform == OCT_MIRROR_X) && (sterm->side == LEFT) && 
                     (counter == nummods) ) ) ) {
               OH_ASSERT (ohGetByTermName (&inst, &octterm, sterm->name) );
               sprintf (terminalname,"%s.%d",sterm->name,counter);
               symCreateFormalTerm (&row,&formalterm,terminalname);
               symImplementFormalTerm (&formalterm,&octterm);
               previousterm = presentterm;
               CALLOC_ERRCHK 
                      (presentterm, 1, struct _terms); 
               if (firstterm==NULL) {
                  firstterm = presentterm;
               } /* if */
               else {
                  previousterm->next = presentterm; 
               } /* else */
               strcpy (presentterm->terminalname,terminalname);
               presentterm->netnumber = POWER_NET;
               strcpy (presentterm->netname,sterm->net);
               strcpy (presentterm->property,sterm->property);
               presentterm->side = sterm->side;
               presentterm->next = NULL;
           } /* if */
       } /* for */ 
    } /* for */

    /*
     * Create Internal Nets for Horizontal Power Lines
     */

    firstnet = NULL;
    /* Search the terms for unique nets and count them */
    for (presentterm = firstterm; presentterm!=NULL; 
	 presentterm = presentterm->next) {
       if (presentterm->netnumber == POWER_NET) {
          if ( (presentterm->side == RIGHT) || (presentterm->side == LEFT) ) {
             found = 0;
             for (searchnet=firstnet; searchnet!=NULL; 
                  searchnet=searchnet->next) {
                if (strcmp(searchnet->term->netname,presentterm->netname)==0) {
                   (searchnet->numterms)++;
                   found = 1;
                }
             }
             if (found == 0) {
                previousnet = presentnet;
                CALLOC_ERRCHK (presentnet, 1, struct _snet);
                if (firstnet == NULL) {
                   firstnet = presentnet;
                } /* if */
                else {
                   previousnet->next = presentnet;
                } /* else */
                presentnet->numterms = 1; 
                presentnet->term = presentterm;
                presentnet->next = NULL; 
             } /* Need to create a new Net */ 
          } /* Valid terminal found */
       } /* POWER terminal */
    } /* for */
    /* 
     * Now go through the nets; look for terminals and attach them to net 
     * after creating them if there are 2 terminals attached to the net
     */
    for (searchnet = firstnet; searchnet!=NULL; searchnet=searchnet->next) {
       if (searchnet->numterms > 1) {
           symCreateNet (&row,&(searchnet->net));
    	   searchnet->net.contents.net.name = searchnet->term->netname;
    	   octModify(&(searchnet->net));
           ohCreatePropStr (&(searchnet->net),&prop,"NETTYPE",
                   searchnet->term->property); 
           for (presentterm=firstterm; presentterm!=NULL; 
                presentterm=presentterm->next) {
              if (presentterm->netnumber == POWER_NET) {
                 if ( (presentterm->side == RIGHT) ||
                      (presentterm->side == LEFT) ) {
                    OH_ASSERT (ohGetByTermName(&row,&octterm, presentterm->terminalname));
                    symAddTermToNet (&octterm,&(searchnet->net));
                 } /* if -add term */ 
              } /* if */
           } /* for */
       } /* if there are at least 2 terminals attached to the net */
    } /* for */

    *final_x_position = next_x_position;
     
    OH_ASSERT (octCloseFacet (&row));    
    /*
     * Make instance of row in facet.
     */ 
    ADSgenInterface (&row);

    inst.type = OCT_INSTANCE;
    inst.contents.instance.master = row.contents.facet.cell;
    inst.contents.instance.view = row.contents.facet.view;
    inst.contents.instance.facet = row.contents.facet.facet;
    inst.contents.instance.version = row.contents.facet.version;
    inst.contents.instance.transform.transformType = OCT_NO_TRANSFORM;
    inst.contents.instance.transform.translation.x = 0;
    inst.contents.instance.transform.translation.y = *y_position;
 
    symCreateInstance (facet,&inst);
    inst.contents.instance.name = name;
    octModify(&inst);
    
    /*
     * Put terminals on nets
     */

    for (presentterm = firstterm; presentterm!=NULL; presentterm= presentterm->next) {
         if (presentterm->netnumber != POWER_NET) {
            if( (ADSnet=ADSsearchADSnetlist(ADSnetlist, presentterm->netnumber)) != NULL ) {
               OH_ASSERT ( ohGetByTermName (&inst,&octterm, presentterm->terminalname) );
               OH_ASSERT ( octAttach (&(ADSnet->net),&octterm) );
            }
         }
    } /* for */

    *y_position += max_y;
} /* ADScreateRow */

static void
  ADSplaceRow(facet, first_module, nameBase, y_position, final_x_position, ADSnetlist, constr)
octObject *facet;
struct     _module *first_module;
char      *nameBase;
int       *y_position;
int       *final_x_position;
_ADSnetlist *ADSnetlist;
int        constr;       /* SYM_VERTICAL and/or SYM_HORIZONTAL */
/*  ============================================================
 *  Parameter <constr> specifies whether to set normal constraints
 *  between the elements on the row. Vertical constraints require
 *  that all elements be aligned with the first one.
 *  Horizontal constraints require that they maintain their current
 *  horiz. offset with respect to the first element.
 *  -- Enrico - Jan. 1991
 */
{
   char                 InstName[20];
   int                  counter = 0;
   int			x_position;
   struct _module 	*mod;
   octTransformType	transform;
   octObject		int_facet; 		 /* Intermediate facet      */
   octBox               BB;
   octObject		instance, firstInst;
   octObject		octterm;
   int			max_y;		         /* Find Maximum Y Height */
   int                  cntr;
   struct _terminal	*term;
   _ADSnetlist	*ADSnet;

   max_y = 0;

   for (cntr=0,mod=first_module; mod != NULL; mod=mod->next, cntr++) {

       /* Detect the real size of every element  */
       findRealBB( mod, &BB );
       mod->xdim = BB.upperRight.x - BB.lowerLeft.x;
       mod->ydim = BB.upperRight.y - BB.lowerLeft.y;

      if (mod->transform == NO) {
	  transform = OCT_NO_TRANSFORM;
	  x_position = (*final_x_position) + mod->xsep;
	  *final_x_position = x_position + mod->xdim; 
      } else  {  /* if (mod->transform == YES) { */
	  transform = OCT_MIRROR_X; 
	  x_position = (*final_x_position) + mod->xsep + mod->xdim;
	  *final_x_position = x_position;
      }
      
      mod->ytrans = *y_position; 

      ohUnpackDefaults (&int_facet,"r",":physical:contents");
      ohUnpackFacetName (&int_facet, mod->name);
      sprintf(InstName, "%s.%d", nameBase, ++counter);  /* Start from 1 */
      instance.type = OCT_INSTANCE;
      instance.contents.instance.master  = int_facet.contents.facet.cell;
      instance.contents.instance.view    = int_facet.contents.facet.view;
      instance.contents.instance.facet   = int_facet.contents.facet.facet;
      instance.contents.instance.version = int_facet.contents.facet.version;
      instance.contents.instance.transform.transformType = transform;
      instance.contents.instance.transform.translation.x = x_position;
      instance.contents.instance.transform.translation.y = mod->ytrans;

      symCreateInstance (facet, &instance);
      instance.contents.instance.name = InstName;
      octModify(&instance);

     /*
      * Now that it has the instance,
      * Find the terminals on them and connect them to the correct net.
      */

      for (term=mod->terminals; term!=NULL; term=term->next) {
         if (strcmp(term->name,"")!=0) {
            if ( (ADSnet=ADSsearchADSnetlist(ADSnetlist,
                  term->netptr->number)) != NULL ) {
               OH_ASSERT ( ohGetByTermName (&instance,&octterm,term->name) );
               OH_ASSERT ( octAttach (&(ADSnet->net),&octterm) );
            } /* if ADSnet */ 
         } /* make sure that there is a term name */
      } /* End end of the terminals */

      if ( mod->ydim > max_y ) {
         max_y = mod->ydim;
      } /* if */

      if(constr & SYM_VERTICAL) {       /* Y constraints  */
	if(cntr == 0)
	  firstInst = instance;
	else
	  symNormalConstraint(SYM_VERTICAL, &firstInst, &instance, 0, 0);
      }
      if(constr & SYM_HORIZONTAL) {       /* X constraints  */
	if(cntr == 0)
	  firstInst = instance;
	else
	  symNormalConstraint(SYM_HORIZONTAL, &firstInst, &instance,
	      x_position - firstInst.contents.instance.transform.translation.x,
	      x_position - firstInst.contents.instance.transform.translation.x);
      }
    } /* for */ 

   *y_position += max_y;

}

ADSplaceConnectors(facet,channel,x_max,y_max,ADSnetlist,leftbag,rightbag,lib)
octObject *facet;
struct _channel *channel;
int x_max;
int *y_max;
_ADSnetlist *ADSnetlist;
octObject *rightbag;
octObject *leftbag;
char *lib;
{
    int i,j,x,y_start;
    struct _terminal 	*term;
    octObject		conn_bag,		/* Connectors Bag */
                        inst_bag;
    char		inst_name[FILENAME_SIZE];
    octObject		inst_facet;		/* Instance Facet */
    octObject		instance;		/* Instance */
    _ADSnetlist  *ADSnet;
    octObject		octterm;
    octObject		formalterm;
    octObject		prop;
    char		termtype[MAX_OCT_NAME];
    char                direction[MAX_OCT_NAME];
    int			prop_flg;

    octObject		*constraintbag;

    sprintf (inst_name,"%s/%s",lib, CONNECTOR);
    OH_ASSERT (ohGetByBagName (facet, &conn_bag, "CONNECTORS"));
    OH_ASSERT (ohGetByBagName (facet, &inst_bag, "INSTANCES"));

    y_start = *y_max;
    for (i=0; i<2; i++) {
       if (i==0) {
          j=LEFT;
          x=0-CONNECTOR_OFFSET;
          constraintbag = leftbag;
       } /* if */
       else {
          j=RIGHT;
          x=x_max+CONNECTOR_OFFSET;
          constraintbag = rightbag;
       } /* else */
       for (term=channel->tlist[j]; term!=NULL; term=term->nxtcht) {
          if (strcmp(term->name,"") != 0) {
             ohUnpackDefaults (&inst_facet,"r",":physical:contents");
             ohUnpackFacetName (&inst_facet,inst_name);
             OH_ASSERT (ohCreateInstance (facet, &instance, term->name,
					  inst_facet.contents.facet.cell,
					  inst_facet.contents.facet.view,
                x, y_start+term->ycoord, OCT_NO_TRANSFORM));
             OH_ASSERT (octAttach (&inst_bag, &instance)); 

            /*
             * Have Instance, attach to constraint bag
             */
             ADSaddConstraint (constraintbag, &instance);
             OH_ASSERT(ohCreatePropStr(
                             &instance,&prop,"PLACEMENT_CLASS","PAD"));
            /*
             * Have Instance, grab terminal, and attach to net.
             */

             if ( (ADSnet=ADSsearchADSnetlist(ADSnetlist,
					      term->netptr->number)) != NULL ) {
	       OH_ASSERT(octGenFirstContent(&instance,OCT_TERM_MASK,&octterm));
	       /* only one terminal */
	       symAddTermToNet (&octterm,&(ADSnet->net));
	       symCreateFormalTerm(facet,&formalterm,term->name);
	       symImplementFormalTerm(&formalterm,&octterm);
	       /*
		* Add the properties while we have the formal terminals
		*/
	       prop_flg = 1;
                switch (term->type) {
		case INPUT 	: strcpy (termtype,"SIGNAL");
		  strcpy (direction,"INPUT");
		  break;
		case OUTPUT	: strcpy (termtype,"SIGNAL");
		  strcpy (direction,"OUTPUT");
		  break;
		case INOUT   : strcpy (termtype,"SIGNAL");
		  strcpy (direction,"INOUT");
		  break;
		  default      : prop_flg = 0;
		} /* Switch */
	       if (prop_flg) {
		 OH_ASSERT(ohCreatePropStr(
			     &formalterm,&prop,"TERMTYPE",termtype));
		 OH_ASSERT(ohCreatePropStr(
			     &formalterm,&prop,"DIRECTION",direction));
	       } /* if */
             } /* if ADSnet */ 

             if ( (y_start+term->ycoord) > *y_max) {
                *y_max = y_start+term->ycoord;
             } /* if */
          } /* if */
       } /* for */
    }

} 

static void
  ADSplaceAll(facet,ADSnetlist,lib, instanceDir, noClockLines)
octObject *facet;
_ADSnetlist *ADSnetlist;
char *lib;
char *instanceDir;
int   noClockLines;
{
    int 	y_position;		/* Place cells row by row */
    int		x_max, x_position;
    octObject	topbag, bottombag, rightbag, leftbag;

    /*
     * Have this routine here, because the only contraints that are needed
     * are for the terminals that are being placed down.
     */
    ADSaddConstraintBags (facet, &topbag, &bottombag, &rightbag, &leftbag); 

    y_position = 0; x_max = 0; 
    ADScreateRow(facet, modlist[OPAMP], "OPAMP", &y_position, &x_position, ADSnetlist, instanceDir, lib);
    if (x_position > x_max) {
       x_max = x_position;
    }
    ADSplaceConnectors(facet, chanlist[BOTTOM],x_max,&y_position,
                       ADSnetlist,&leftbag,&rightbag,lib);
    x_position = 0; y_position += Y_OFFSET;
    ADSplaceRow(facet, modlist[CBLK], "CBLK", &y_position, &x_position, ADSnetlist, 0);
    if (x_position > x_max) {
       x_max = x_position;
    }
    ADSplaceConnectors(facet, chanlist[TOP],x_max,&y_position,
                       ADSnetlist,&leftbag,&rightbag,lib);
    x_position = 0; y_position += Y_OFFSET;  
    ADScreateRow (facet, modlist[SWCH], "SWCH",
		 &y_position, &x_position, ADSnetlist, instanceDir, lib);
    if (x_position > x_max) {
       x_max = x_position;
    }
    ADSplaceConnectors(facet,chanlist[CLKCH],x_max,&y_position,
		       ADSnetlist,&leftbag,&rightbag,lib);
    if(! noClockLines ) {
	VERBOSE( printf( "Routing the clock lines\n") );
	ADSgenClockWires(facet, instanceDir);
    }
    /*
     * Now x_max contains maximum value of x
     * y1,y2,y3 contain the bottom y coordinates of the three channels.
     */ 

}

ADSdisplayChannel (channel,name)
struct _channel *channel;
char *name;
{
   int 			i,j;
   char 		js[10];
   struct _terminal 	*term;

   if (channel==NULL) {
      printf ("Channel %s is NULL\n",name); 
   }
   else {
      printf ("DISPLAYING CHANNEL: %s\n",name);
      for (i=0; i<4; i++) {
         j = 0; strcpy (js, "NOTHING");
         switch (i) {
            case 0 : j=TOP; strcpy (js,"TOP"); break;
            case 1 : j=BOTTOM; strcpy (js,"BOTTOM"); break;
            case 2 : j=RIGHT; strcpy (js,"RIGHT"); break;
            case 3 : j=LEFT; strcpy (js,"LEFT"); break;
         } /* switch */
         printf ("Side of Channel: %8s     Number of Terms: %d\n",
      		   js, channel->numterms[j]);
         for (term=channel->tlist[j]; term!=NULL; term=term->nxtcht) {
            printf 
               ("Mod(%d): %21s  Terminal c (%4d,%4d)  chc (%4d,%4d)  l:%d\n",
                term->module->type,term->module->name, 
	   	term->xcoord,term->ycoord,term->chxcoord,term->chycoord,
		term->chrlayer);
            printf ("    name: %s\n",term->name);
         }
      }
   } /* else */  
} /* ADSdisplayChannel */

static void
  ADScreateCBLKterms()
/* Creates Formal Terminals on Capacitor Leaf Cells based on information
 * in the data base.
 * Opens the CBLK facets directly.
 * Names the terminals in order of linked list in data base.
 * 1st one = module->terminals.  ("term_n")
 * Store the terminal names in term->name.  This field does not
 * seem to be used by adore.
 * -hcc
 */
{
    struct _module *mod;
    struct _terminal *term;
    int counter;

    octObject facet;
    octObject met_layer,poly_layer;
    octObject formal_term;
    octObject Net;          /*  Net  */
    octObject box;
    int       width;
    octObject *layer;
    int	      llx,lly,urx,ury;
    char NetName[40];       /*  Net name */

    for (mod=modlist[CBLK]; mod!=NULL; mod=mod->next) {
       counter = 0;
       ohUnpackDefaults (&facet, "a", ":physical:contents");
       ohUnpackFacetName (&facet, mod->name);
       OH_ASSERT (octOpenFacet(&facet));
       OH_ASSERT (ohGetByLayerName (&facet, &met_layer, metal_L));
       OH_ASSERT (ohGetByLayerName (&facet, &poly_layer, botpoly_L));
       for (term=mod->terminals; term!=NULL; term=term->next) {
	   if(term->chrlayer == NOT_ASSIGNED) continue;
	   sprintf (term->name,"term_%d",counter);
	   OH_ASSERT (ohCreateTerm (&facet,&formal_term,term->name));
	   if (term->chrlayer==1) {
	       width = polywidth; 
	       layer = &poly_layer;
	   } else {
	       width = metwidth;
	       layer = &met_layer;
	   } /* else term */
	   llx = term->xcoord - width / 2;   LAMBDA_ROUND(llx);
	   urx = term->xcoord + width / 2;   LAMBDA_ROUND(urx);
	   lly = term->ycoord - width / 2;   LAMBDA_ROUND(lly);
	   ury = term->ycoord + width / 2;   LAMBDA_ROUND(ury);
	   if (MOSAICO_HELP) { /* Make Terminals longer */ 
	       if (term->channel->name == TOP) {
		   ury += MOSAICO_OFFSET; 
	       }   
	       if (term->channel->name == BOTTOM) {
		   lly -= MOSAICO_OFFSET; 
	       }   
	   }
	   OH_ASSERT (ohCreateBox(&formal_term,&box,llx,lly,urx,ury));
	   OH_ASSERT (octAttach(layer,&box));
	   
	   sprintf(NetName, "net_%d", term->netptr->number);    /* Net name  */
	   OH_ASSERT (ohGetOrCreateNet(&facet, &Net, NetName)); /* Net  */
	   OH_ASSERT (octAttach(&Net, &formal_term));           /* Attach */
	   
	   counter++;
       } /* for term */ 
       OH_ASSERT (octCloseFacet (&facet));
    } /* for mod */ 

}

/*
 * ADSdisplayNets
 * starts at struct _net *netlist, and finds all the nets and all of
 * terminals.
 */

ADSdisplayNets ()
{
    struct _net *net;
    struct _terminal *term;

    for (net=netlist; net!=NULL; net=net->next) {
       printf ("NET name=%s  number=%d\n",net->name,net->number);
       for (term=net->firstnett; term!=NULL; term=term->nxtnett) {
          printf ("   TERM name=%s   module name=%s   (%d,%d)\n",
             term->name, term->module->name, 
             term->xcoord, term->ycoord); 
       }
    } /* for: net search */

}


ADsymbolic (outputcell,lib,instanceDir, noClockLines, noWells)
char *outputcell;
char *lib;			/* only here for the com2.4x4 connector */
char *instanceDir;
int   noClockLines;
int   noWells;
{
    octObject		 facet;		/* Output Symbolic Cell */
    octObject            prop;
    _ADSnetlist	*ADSnetlist;	/* netlist with OCT information */
    char wellType  = 'n';
    char *wellBias = "VDD";
    octBegin();

    ADScreateCBLKterms(); 
    ADScreateCBLKifaces();

    ohUnpackDefaults (&facet,"w",":symbolic:contents");
    ohUnpackFacetName (&facet,outputcell);
    symOpenFacet (&facet, "w", technName );

    ADScreateADSnetlist (&facet,&ADSnetlist);

    OH_ASSERT(ohCreatePropStr (&facet, &prop, "CELLCLASS",  "MODULE"));
    ADSplaceAll(&facet,ADSnetlist,lib,instanceDir, noClockLines);

    if(! noWells ) {
	VERBOSE( printf("Generate %c-well under the capacitors\n", wellType) );
	adWells( &facet, wellType, wellBias, instanceDir );
    }

    OH_ASSERT ( octCloseFacet (&facet) );
    ADSgenInterface (&facet); 

    octEnd();
}
