/* Evaluate the dimension of capacitor blocks for Adore2 */
/* Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli */
/* last change: 7/9/87 */

/* defined: CBLKfindDim, CBLKfindDTA, CBLKfindCTA */

/*
 * Modifications to Adore
 * Henry Chang (henryc@ic.Berkeley.EDU)
 * Fall, 1990
 */

#define C_FILE_NAME "CBLKfindDim.c"

#include "ADglobals.h"
#include "TEKtech.h"
#include <stdio.h>
#include <math.h>


/* This function estimates the optimum height of each block */
/* Please refer to notes of 12/13/86 */

static int maxydim;                /* maximum height of all capacitor blocks */
static int maxChipWidth;           /* maximim width of allblocks */

static int maxoah;
static int maxswh;

static int H, W, Ts;                                      /* constant dimensions */
static int radius;                /* for searching the optimum discrete solution */
static int extend;                       /* extension of the unit capacitor side */
FILE *fp1, *fp2, *fp3;
double ptime(), stime, etime;
int abs();

static int numtracks[MAXNUMCH];

CBLKfindDim()
{
  FILE *fopen();

  int Wamp, Wswch;		  /* width of the opamp and switch arrays */

  double CBLKfindCTA();
  double CBLKfindDTA();
  double TAnewa, TAnewb, TAb;	    /* continuous total areas for bracket */
  double ha, hb, hnewa, hnewb, interval;
  double Fibo = 1. - .61804;	                /* used for golden search */
  double cwidth, cheight;
  double DTA, optDTA;
  int dheight, dhmin, dhmax;
  int dhopt,dwopt;	   /*discrete total area,optimal height and width */
  int dhexact; 

  struct _module *tmpmod;
	
  H = 0;
  W = 0;
  Ts = 0;
  radius = 400 * resolution/2;

  extend = 2 * botextop + botbotsp;

  if ( H == 0 ) {
    maxoah = 0;
    for ( tmpmod = modlist[OPAMP]; tmpmod != NULL; tmpmod = tmpmod->next) {
      if ( tmpmod->ydim > maxoah ) {
	maxoah = tmpmod->ydim;
      }
    }

    maxswh = 0;
    for ( tmpmod = modlist[SWCH]; tmpmod != NULL; tmpmod = tmpmod->next) {
      if ( tmpmod->ydim > maxswh ) {
	maxswh = tmpmod->ydim;
      }
    }

    /* find estimate of H */
    numtracks[TOP] = 3;
    numtracks[BOTTOM] = 2;
    H = maxoah + maxswh + 
      minctctsp * ( numtracks[TOP] + 1 + numtracks[BOTTOM] + 1 );
  }

  if (W == 0) {
    Wamp = 0;
    for ( tmpmod = modlist[OPAMP]; tmpmod != NULL; tmpmod = tmpmod->next) {
      Wamp += tmpmod->xdim ;
    }
    Wamp += (Numinteg - 1) * oaxsep;

    Wswch = 0;
    for ( tmpmod = modlist[SWCH]; tmpmod != NULL; tmpmod = tmpmod->next) {
      Wswch += tmpmod->xdim;
    }

    W = max(Wamp, Wswch);
  }

  if ( Ts == 0 ) {
    Ts = (Numcblk - 1)  *  blkxsep  +
      2 * (minoverhang + max(polywidth,metwidth));
  }

   DEBUG (printf("H=%d, W=%d, Ts=%d\n",H, W, Ts));
   DEBUG (printf("Minside=%d,extend=%d\n\n", Minside,extend));


  /* golden search for one dimensional optimization */
  stime = ptime();
  ha = Minside + extend;
  /* evaluate hb so that w >= W - Ts */
  hb = ha;
  cwidth = 1.e99;
  while ( cwidth > (double) ( W - Ts) ) {
    CBLKfindCTA(hb,&cwidth,&cheight);
    if ( hb > cheight ) {
      hb = cheight;
      break;
    }
    hb = hb * cwidth / (double)(W-Ts) + 10;
  }
  
  DEBUG(printf("hb=%g width=%g\n",hb,cwidth));

  interval = hb - ha;
  TAb = CBLKfindCTA(hb,&cwidth,&cheight);
  while ( interval > .1 * resolution ) {
    interval = hb - ha;
    hnewa = ha + interval * Fibo;
    hnewb = hb - interval * Fibo;
    TAnewa = CBLKfindCTA(hnewa,&cwidth,&cheight);
    TAnewb = CBLKfindCTA(hnewb,&cwidth,&cheight);
    if ( TAnewb <= min(TAnewa, TAb) || TAnewb > TAb ) {
      ha = hnewa;
    }
    else {
      hb = hnewb;
      TAb = CBLKfindCTA(hb,&cwidth,&cheight);
    }
  }
  
  DEBUG(printf("from golden search h= %g, minTA=%g\n",hb, TAb));
  
  etime = ptime();
  DEBUG(printf("total time for golden search=%g\n",etime- stime));

  /* now find optimal discrete solution */
  stime = ptime();
  optDTA = 1.e99;
  dhmin = max((int)hb - radius, Minside + extend);
  dhmin = Round_Int(dhmin, resolution/2, CEIL );
  dhmax = (int) hb + radius;
  dhmax = Round_Int(dhmax, resolution/2, CEIL);
  for ( dheight = dhmin; dheight <= dhmax; dheight += (resolution / 2) ) {
    DTA = CBLKfindDTA(dheight, &dwopt, &dhexact);
    if ( DTA < optDTA ) {
      optDTA = DTA;
      dhopt = dheight;
    }
  }
  /* this is used to update the internal data structure */
  optDTA = CBLKfindDTA(dhopt, &dwopt, &dhexact);
  maxydim = Round_Int(dhopt, resolution/2, CEIL);

  maxChipWidth = max(dwopt + Ts, W);
  maxChipWidth = Round_Int(maxChipWidth, resolution/2, CEIL);
  if ( Numsw > 1 ) {
    swxsep = (maxChipWidth - Wswch) / (Numsw - 1);
  }
  if ( Numinteg > 1 ) {
    oaxsep += (maxChipWidth - Wamp) / (Numinteg - 1);
  }
  if ( Numcblk > 1 ) {
    blkxsep += (maxChipWidth - dwopt - Ts) / (Numcblk - 1);
  }
  swxsep = Round_Int(swxsep, resolution, CEIL);
  oaxsep = Round_Int(oaxsep, resolution, CEIL);
  blkxsep = Round_Int(blkxsep, resolution, CEIL);


  DEBUG(printf("radius=%d, dhopt=%d, optDTA=%g, dwopt=%g\n",radius,dhopt,
	optDTA, dwopt));
  etime = ptime();
  DEBUG(printf("total time for discrete solution=%g\n",etime- stime));
}

/* find continuous total area */
double CBLKfindCTA(heff, pwidth, pheight)
double heff, *pwidth, *pheight;
{
  int counter;
  double minnumx, maxnumx, optnumx, minnumy, maxnumy, optnumy;
  double unitA, optunith, optunitw;
  double minxdim, optydim; /* holds the minimum value for a given block */
  double ydimceil;
  double totxdim; /* this should be compared with Weff */
  double TA;

  struct _module *tmpmod;
  struct _cblk *tmpcblk;
  
  totxdim = 0;
  ydimceil = 0;
  
  for ( tmpmod = modlist[CBLK]; tmpmod != NULL; tmpmod = tmpmod->next ) {
    tmpcblk = tmpmod->info.cblk;
    minnumx = max(tmpcblk->ncontc, tmpcblk->nswitc);
    maxnumx = tmpcblk->totalsum;
    minnumy = 1;
    maxnumy = maxnumx / minnumx;
    unitA = tmpcblk->unitA;
    
    optunith = pow(unitA,.5);
    optnumy = (heff + botbotsp) / ( extend + optunith ) ;
    if ( optnumy < minnumy ) {
      optnumy = minnumy;
      optunith = (heff + botbotsp)/optnumy - extend;
    }
    else if ( optnumy > maxnumy ) {
      optnumy = maxnumy;
      optunith = min((heff+botbotsp)/optnumy - extend, unitA/Minside);
    }
    optnumx = maxnumx / optnumy;
    optunitw = unitA/ optunith;
    minxdim = optnumx * ( optunitw + extend ) - botbotsp;
    optydim = optnumy * ( optunith + extend ) - botbotsp;
    
    totxdim += minxdim;
    
    if ( optydim > ydimceil ) {
      ydimceil = optydim;
    }
    
  }
  *pwidth = totxdim;
  *pheight = ydimceil;
  /* evaluate total area */
  TA = (H + ydimceil) * max(W, totxdim + Ts);
  return(TA);
}

/* This function estimates the optimum Total Area for a given height */
/* Please refer to notes of 7/23/86 */
/* find discrete total area */
double
CBLKfindDTA( heff, pwidth, pheight)
int heff, *pwidth, *pheight;
{
  int counter;
  double TA;
  int unitA, unith, unitw, xdim, ydim; /* temporary storage variables */
  int numx, minnumx, maxnumx, optnumx; /* # of unit caps in x direction */
  int numy, minnumy, maxnumy, optnumy; /* # of unit caps in y direction */
  int optunith; /*optimum value of unit height */
  int optunitw, optydim;
  int minxdim; /* holds the minimum value for a given block */
  int totxdim; /* this should be compared with Weff */
  int ydimceil;
  
  struct _module *tmpmod;
  struct _cblk *tmpcblk;
  
  totxdim = 0;
  ydimceil = 0;
  
  for ( tmpmod = modlist[CBLK]; tmpmod != NULL; tmpmod = tmpmod->next ) {
    tmpcblk = tmpmod->info.cblk;
    minnumx = max(tmpcblk->ncontc, tmpcblk->nswitc);
    maxnumx = tmpcblk->totalsum;
    minnumy = 1;
    maxnumy = ceil ( (double)maxnumx / (double)minnumx );
    unitA = tmpcblk->unitA;
    
    minxdim = MAXINT;
    optydim = MAXINT;
    optnumx = 0;
    for ( numy= maxnumy; numy >=minnumy; numy--) {
      
      unith = (heff + botbotsp)/numy - extend;
      unith = Round_Int(unith, resolution, FLR);
      
      if ( unith < Minside ) {
	continue;
      }
      
      unitw = ceil((double)unitA/ (double) unith);
      unitw = Round_Int(unitw, resolution, CEIL);
      /* the following is necessary to find the
	 minimum unith for a given unitw */
      /*
	unith = ceil((double)unitA/ (double) unitw);
	unith = Round_Int(unith, resolution, RND);
	*/
      
      if ( unitw < Minside ) {
	unitw = Minside;
	unith = ceil((double)unitA/ (double) unitw);
	if ( unith < Minside ) {
	  printf("AREA < MINSIDE**2\n");
	  exit ( -1 ) ;
	}
      }
      
      numx = ceil((double)maxnumx / (double)numy);
      xdim = numx * ( unitw + extend ) - botbotsp;
      ydim = numy * ( unith + extend ) - botbotsp;
      
      
      
      if ( ydim <= heff ) {
	if ( xdim < minxdim ||(xdim == minxdim && ydim < optydim)) {
	  minxdim = xdim; optydim = ydim;
	  optnumy = numy; optunith = unith;
	  optnumx = numx; optunitw = unitw;
	}
      }
      if ( unitw == Minside ) {
	/* as we decrease numy, unith goes up,
	   unitw goes down */
	break;
      }
    }
    totxdim += minxdim;
    
    tmpcblk->unith = optunith;
    tmpcblk->unitw = optunitw;
    tmpcblk->numrow = optnumy;
    tmpcblk->numcol = optnumx;
    
    tmpmod->xdim = minxdim;
    tmpmod->ydim = optydim;
    
    if ( optydim > ydimceil ) {
	ydimceil = optydim;
    }
  }
  *pwidth = totxdim;
  *pheight = ydimceil;
  /* evaluate total area */
  TA = (double)(H + ydimceil) * (double) max(W, totxdim + Ts);
  return(TA);
}
