/* finds the location of capacitor terminals for a cblk for adore2 */

/* Author:Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli
 * last change : 7/15/87 
 * Henry Chang (henryc@ic.Berkeley.EDU)  -  Fall, 1990
 * Last revised: March 1992  -  Enrico Malavasi */

#define C_FILE_NAME "CBLKgenTerms.c"

#include "TEKtech.h"
#include "CBLKglobals.h"
#include "oct.h"
#include "oh.h"
#include <stdio.h>

octObject array_contents;
static int deltax, deltay;

static int maxydim;

static void
  CBLKequivAmpIO( cblk, writeoct )
struct _cblk *cblk;
int writeoct;
/* generate the extensions of opamp input and outputs to the top channel */
{
    int rowNo, colNo;
    int lowleftx, lowlefty, uprightx, uprighty;
    int nlowleftx, nlowlefty, nuprightx, nuprighty;
    int layer;
    struct _terminal *equivTerm;
    
    
    /*** In the master of the cblk, equivalent output is on the left side,
      and equivalent input is on the right side. (the integrating capacitor
      is the first capacitor of the bottom row of the cblk)*/
    
    /*** On the left side, start from the top row, as soon as we find 
      the first integrating unit capacitor we stop */
    
    if ( ucap_matrix[cblk->numrow][1] == cblk->integcap ) {
	/*  Upper side - POLY vertical stub */
	lowleftx = 0;
	lowlefty = cblk->numrow * deltay - botbotsp;
	uprightx = lowleftx + polywidth;
	uprighty = maxydim;
	if ( lowlefty < uprighty ) {
	    if ( writeoct == YES ) {
		makebox(&array_contents,botpoly_L,lowleftx,lowlefty,uprightx,uprighty);
	    }
	}
	layer = 1;
    } else {
	for ( rowNo = cblk->numrow - 1;rowNo >= 1; rowNo--) {
	    if ( ucap_matrix[rowNo][1] == cblk->integcap ) {
		break;
	    }
	}
	if ( ucap_matrix[rowNo][1] != cblk->integcap ) {
	    errRaise(C_FILE_NAME, -1, "CBLKequivAmpIO;1: internal error");
	}
	/* Left side - POLY horizontal stub */
	uprightx = 0;
	uprighty = rowNo * deltay - botbotsp;
	lowleftx = uprightx - minoverhang;
	lowlefty = uprighty - polywidth;
	if ( writeoct == YES ) {
	    makebox(&array_contents,botpoly_L,lowleftx, lowlefty,uprightx,uprighty);
	}
	/* Now make the run to top of the array */
	if ((maxydim - uprighty) <= minctctsp ) {
	    /*  POLY vertical run to the top of the array */
	    lowleftx -= polywidth;
	    uprightx = lowleftx + polywidth;
	    uprighty = maxydim ;
	    if ( lowlefty != uprighty ) {
		if ( writeoct == YES ) {
		    makebox(&array_contents,botpoly_L, lowleftx,lowlefty,uprightx,uprighty);
		}
	    }
	    layer = 1;
	} else {   /*  MET1 vertical run to the top of the array */
	    nlowleftx = lowleftx - ( metwidth + mincontd) / 2;  
	    nlowlefty = lowlefty + (polywidth - mincontd) / 2;
	    nuprightx = nlowleftx + mincontd;
	    nuprighty = nlowlefty + mincontd;
	    if ( writeoct == YES ) {
		makebox(&array_contents,con_L,nlowleftx,nlowlefty,nuprightx,nuprighty);
	    }
	    nlowleftx -= polyovcon;
	    nlowlefty -= polyovcon;
	    nuprightx += polyovcon;
	    nuprighty += polyovcon;
	    if ( writeoct == YES ) {
		makebox(&array_contents,botpoly_L,nlowleftx,nlowlefty,nuprightx,nuprighty);
	    }
	    nlowleftx -= metovcont- polyovcon;
	    nlowlefty -= metovcont- polyovcon;
	    nuprightx += metovcont- polyovcon;
	    nuprighty += metovcont- polyovcon;
	    if ( writeoct == YES ) {
		makebox(&array_contents,metal_L,nlowleftx,nlowlefty,nuprightx,nuprighty);
	    }
	    lowleftx -= metwidth;
	    uprightx = lowleftx + metwidth;
	    uprighty = maxydim ;
	    if ( writeoct == YES ) {
		makebox(&array_contents,metal_L,lowleftx,lowlefty,uprightx,uprighty);
	    }
	    layer = 2;
	}
    }
    
    equivTerm = cblk->eqbotpt;
    equivTerm->xcoord = (lowleftx + uprightx) / 2;
    equivTerm->ycoord = uprighty;
    equivTerm->chrlayer = layer;
    
    /*** On the right side start from the top row, as soon as we find 
      the first continuous unit capacitor we stop */

    if(cblk->numrow == 1) {    /*  added by Enrico - 10/26/1990 */
	for( colNo=1 ; colNo<=cblk->numcol ; colNo++ ) {
	    if ( ucap_matrix[cblk->numrow][colNo]->type == CONT &&
		ucap_flags[cblk->numrow][colNo] != VNUC ) {
		break;
	    }
	}
	if( colNo <= cblk->numcol ) {
	    lowleftx = colNo * deltax - (deltax+botbotsp+metwidth)/2;  LAMBDA_ROUND(lowleftx);
	    lowlefty = deltay - botextop; LAMBDA_ROUND(lowlefty);
	    uprightx = lowleftx + metwidth;
	    uprighty = maxydim;
	    if ( writeoct == YES ) {
		makebox(&array_contents,metal_L,lowleftx,lowlefty, uprightx,uprighty);
	    }
	    equivTerm = cblk->eqtoppt;
	    equivTerm->xcoord = (lowleftx + uprightx) / 2;
	    equivTerm->ycoord = uprighty;
	    equivTerm->chrlayer = 2;   
	}
	return;
    } else {
	for ( rowNo = cblk->numrow; rowNo >=1; rowNo--) {
	    if ( ucap_matrix[rowNo][cblk->numcol] != NULL ) {
		if ( ucap_matrix[rowNo][cblk->numcol]->type == CONT &&
		    ucap_flags[rowNo][cblk->numcol] != VNUC ) {
		    break;
		}
	    } else
	      VERBOSE (printf
		       ("CBLKequivAmpIO: Warning: ucap_matrix[%d][%d] is NULL.\n",
			rowNo,cblk->numcol));
	}
	if ( ucap_matrix[rowNo][cblk->numcol] != NULL ) {
	    if(ucap_matrix[rowNo][cblk->numcol]->type > 0) {
		/* if/else added by Enrico  10/28/90 */
		if ( ucap_matrix[rowNo][cblk->numcol]->type != CONT ||
		    ucap_flags[rowNo][cblk->numcol] == VNUC ) {
		    errRaise(C_FILE_NAME, -1, "CBLKequivAmpIO;2: internal error\n");
		}
	    } else {
		rowNo = 1;
	    }
	} else {
	    VERBOSE (printf
		     ("CBLKequivAmpIO: Warning: ucap_matrix[%d][%d] is NULL.\n",
		      rowNo,cblk->numcol));
	    rowNo = 1;
	}
    }
    
    if ( uctype == 1 ) {
	lowleftx = cblk->numcol * deltax -( botextop + botbotsp ); LAMBDA_ROUND(lowleftx);
	lowlefty = (rowNo - 1) * deltay + botextop + (cblk->unith - metwidth)/2; LAMBDA_ROUND(lowlefty);
	uprightx = lowleftx + botextop + minoverhang;
	uprighty = lowlefty + metwidth;
	if ( writeoct == YES ) {
	    makebox(&array_contents,metal_L,lowleftx,lowlefty, uprightx,uprighty);
	}
    }
    else if ( uctype == 2 ) {
	lowleftx = cblk->numcol  * deltax -( botextop + botbotsp );
	lowlefty = (rowNo - 1) * deltay + botextop + (cblk->unith-polywidth)/2 ;
	uprightx = lowleftx + botextop + minoverhang;
	uprighty = lowlefty + polywidth;
	if ( writeoct == YES ) {
	    makebox(&array_contents,toppoly_L,lowleftx,lowlefty, uprightx,uprighty);
	}
	/* make the contact to metal */
	nlowleftx = uprightx + ( metwidth - mincontd) / 2;  
	nlowlefty = lowlefty + (polywidth - mincontd) / 2;
	nuprightx = nlowleftx + mincontd;
	nuprighty = nlowlefty + mincontd;
	if ( writeoct == YES ) {
	    makebox(&array_contents,con_L,nlowleftx,nlowlefty, nuprightx,nuprighty);
	}
	nlowleftx -= polyovcon;
	nlowlefty -= polyovcon;
	nuprightx += polyovcon;
	nuprighty += polyovcon;
	if ( writeoct == YES ) {
	    makebox(&array_contents,toppoly_L,nlowleftx,nlowlefty, nuprightx,nuprighty);
	}
	nlowleftx -= metovcont- polyovcon;
	nlowlefty -= metovcont- polyovcon;
	nuprightx += metovcont- polyovcon;
	nuprighty += metovcont- polyovcon;
	if ( writeoct == YES ) {
	    makebox(&array_contents,metal_L,nlowleftx,nlowlefty, nuprightx,nuprighty);
	}
	
    }
    /* now make the run to top of the array */
    lowleftx = uprightx;
    /* lowlefty remains the same */
    uprightx = lowleftx + metwidth;
    uprighty = maxydim ;
    if ( writeoct == YES ) {
	makebox(&array_contents,metal_L,lowleftx,lowlefty, uprightx,uprighty);
    }
    
    equivTerm = cblk->eqtoppt;
    
    equivTerm->xcoord = (lowleftx + uprightx) / 2;
    equivTerm->ycoord = uprighty;
    equivTerm->chrlayer = 2;
}

CBLKgenTerms( cblk, writeoct, instances )
struct _cblk *cblk;
int writeoct;
char *instances;
{
    char cblkname[100], cblkdirname[100];
    char *CBLKmkName(), *CBLKpushDir();
    
    deltax = cblk->unitw + 2 * botextop + botbotsp;
    deltay = cblk->unith + 2 * botextop + botbotsp;
    
    /******* temporary assignment **********/
    maxydim = cblk->module->ydim;
    
    strcpy(cblkname, CBLKmkName(cblk->opamp->name, cblk->in_polarity, "CBLK"));
    strcpy(cblkdirname, CBLKpushDir(instances, cblkname));
    
    DEBUG (printf
	   ("Opening Facet with name %s with \"a\" option...\n",cblkdirname));
    
    openfacet(&array_contents, cblkdirname, "physical", "contents", "a");
    
    DEBUG (printf ("Number of rows = %d,  MAXNUMROW = %d\n",cblk->numrow, MAXNUMROW));
    
    CBLKequivAmpIO( cblk, writeoct );
    
    OH_ASSERT(octCloseFacet(&array_contents));
}
