/* 
 * This contains the programs for allocating memory
 * and initializing various layout elements for the program for automatic
 * layout of switched capacitor circuits 
 * Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli 
 * last change: 7/7/87
 * Modifications to Adore
 * Henry Chang (henryc@ic.Berkeley.EDU) - Fall, 1990
 */

/*
 * defined: INPcrtIOmod INPcrtInteg, INPcrtSw, INPcrtCap, INPcrtTerm,
 *          INPinitMod INPaddList, INPcrtNet, INPcrtLibelm, INPaddLib
 */ 

#define C_FILE_NAME "INPcrtElm.c"

#include <stdio.h>
#include "errtrap.h"
#include "utility.h"

#include "ADglobals.h"


/* Allocate memory and initialize dummy input and output modules */
INPcrtIOmod()
{
    struct _module *tmpmod;
    
    /* each dummy module has the list of overall input or overall output terms*/
    
    /* allocate memory for input block */
    tmpmod = ALLOC( struct _module, 1 );
    INPinitMod(tmpmod);
    tmpmod->type = INMOD;
    modlist[INMOD] = tmpmod;
    
    /* output block */
    tmpmod = ALLOC( struct _module, 1 );
    INPinitMod(tmpmod);
    tmpmod->type = OUTMOD;
    modlist[OUTMOD] = tmpmod;
}

/* Allocates memory and initializes opamp and capacitor block */
INPcrtInteg(name, netptr1, netptr2, netptr3, netptr4, libptr)
char *name; /* name of the opamp */
/* outp outm inp inm */
struct _net *netptr1, *netptr2, *netptr3, *netptr4; 
struct _lib *libptr; /* file location of the library element */
{
    struct _module *ampmod;
    struct _module *cblkmod;
    struct _cblk *tmpcblk;
    struct _opamp *tmpamp;
    struct _terminal *INPcrtTerm();
    
    int counter;
    
    
    Numinteg++;
    /*** initialize the opamp */
    ampmod = ALLOC( struct _module, 1 );
    INPinitMod(ampmod); 
    
    ampmod->type = OPAMP;
    tmpamp = ALLOC( struct _opamp, 1 );
    strcpy(tmpamp->name , name);
    /*** allocate memory for the terminals */
    tmpamp->out[PLUS] = INPcrtTerm(netptr1, chanlist[BOTTOM], OUTPUT, ampmod, NULL);
    tmpamp->out[MINUS] = INPcrtTerm(netptr2, chanlist[BOTTOM], OUTPUT, ampmod, NULL);
    tmpamp->inp[PLUS] = INPcrtTerm(netptr3, chanlist[BOTTOM], INPUT, ampmod, NULL);
    tmpamp->inp[MINUS] = INPcrtTerm(netptr4, chanlist[BOTTOM], INPUT, ampmod, NULL);
    ampmod->info.opamp = tmpamp;
    tmpamp->module = ampmod;
    
    /*** add the module to the list of opamps of the same kind */
    INPaddLib(ampmod, libptr);
    
    /*** add the created oapmp to the list of opamps */
    INPaddList(ampmod, OPAMP);
    
    /*** there could be two capacitor blocks associated with one opamp */
    /*** one for plus and one for minus inputs (fully differential) */
    for ( counter=1; counter <=2; counter++) {
	/* initialize a block for each not grounded input */
	if ( (counter == 1 && netptr3->type != GROUND) ||
	    (counter == 2 && netptr4->type != GROUND) ) {
	    Numcblk++;
	    /*** initialize the capacitor block */
	    cblkmod = ALLOC( struct _module, 1 );
	    INPinitMod(cblkmod); 
	    cblkmod->type = CBLK;
	    
	    tmpcblk = ALLOC( struct _cblk, 1 );
	    tmpcblk->module = cblkmod;
	    cblkmod->info.cblk = tmpcblk;
	    
	    tmpcblk->unitA = NOT_ASSIGNED;
	    tmpcblk->unith = NOT_ASSIGNED;
	    tmpcblk->unitw = NOT_ASSIGNED;
	    tmpcblk->numrow = NOT_ASSIGNED;
	    tmpcblk->numcol = NOT_ASSIGNED;
	    tmpcblk->topsum = tmpcblk->botsum = tmpcblk->totalsum =NOT_ASSIGNED;
	    
	    tmpcblk->ncontc = NOT_ASSIGNED;
	    tmpcblk->contcaps = NULL;
	    tmpcblk->integcap = NULL;
	    tmpcblk->nswitc = NOT_ASSIGNED;
	    tmpcblk->swcaps = NULL;
	    
	    tmpcblk->in_polarity = NOT_ASSIGNED;
	    
	    /*** add the created cblk to the list of cblks */
	    INPaddList(cblkmod, CBLK);
	    
	    /* complete the link between the cblk and its opamp */
	    tmpcblk->opamp = tmpamp;
	    if ( counter == 1 ) {
		tmpamp->cblk[PLUS] = tmpcblk;
		tmpcblk->in_polarity = PLUS;
		
		tmpcblk->eqbotpt = INPcrtTerm(netptr2, chanlist[TOP],
					      OUTPUT, cblkmod, NULL);
		tmpcblk->eqbotpt->chrlayer = 1;
		tmpcblk->eqbotpt->chloc = BOTTOM;
		
		tmpcblk->eqtoppt = INPcrtTerm(netptr3, chanlist[TOP],
					      INPUT, cblkmod, NULL);
		tmpcblk->eqtoppt->chrlayer = 2;
		tmpcblk->eqtoppt->chloc = BOTTOM;
	    }
	    if ( counter == 2 ) {
		tmpamp->cblk[MINUS] = tmpcblk;
		tmpcblk->in_polarity = MINUS;
		tmpcblk->eqbotpt = INPcrtTerm(netptr1, chanlist[TOP],
					      OUTPUT, cblkmod, NULL);
		tmpcblk->eqbotpt->chrlayer = 1;
		tmpcblk->eqbotpt->chloc = BOTTOM;
		
		tmpcblk->eqtoppt = INPcrtTerm(netptr4, chanlist[TOP],
					      INPUT, cblkmod, NULL);
		tmpcblk->eqtoppt->chrlayer = 2;
		tmpcblk->eqtoppt->chloc = BOTTOM;
	    }
	}
    }
    
    /*** each net connected to a terminal of this opamp belongs to the opamp */
    if ( netptr1->type != GROUND ) {
	netptr1->type = OUT;
	netptr1->module = ampmod;
    }
    if ( netptr2->type != GROUND ) {
	netptr2->type = OUT;
	netptr2->module = ampmod;
    }
    if ( netptr3->type != GROUND ) {
	netptr3->type = IN;
	netptr3->module = ampmod;
    }
    if ( netptr4->type != GROUND ) {
	netptr4->type = IN;
	netptr4->module = ampmod;
    }
    /* the inp and inm nets are sensitive */
    netptr3->sensitive = YES;
    netptr4->sensitive = YES;
}

INPcrtSw(name, amp1net, clk1net, capnet, clk2net, amp2net, libptr)
char *name;
struct _net *amp1net, *clk1net, *capnet, *clk2net, *amp2net;
struct _lib *libptr; /* file location of the library element */
{
    struct _module *swmod;
    struct _swch *tmpsw;
    struct _terminal *tmpterm, *INPcrtTerm();
    
    Numsw++;
    swmod = ALLOC( struct _module, 1 );
    INPinitMod(swmod);
    
    swmod->type = SWCH;
    tmpsw = ALLOC( struct _swch, 1 );
    strcpy(tmpsw->name , name);
    tmpsw->type =  NOT_ASSIGNED;
    swmod->info.swch = tmpsw;
    tmpsw->module = swmod;
    
    tmpsw->amp1term = INPcrtTerm(amp1net, chanlist[TOP], INOUT, swmod, NULL);
    tmpsw->clk1term = INPcrtTerm(clk1net, chanlist[CLKCH], CLOCK, swmod, NULL);
    tmpsw->capterm = INPcrtTerm(capnet, chanlist[TOP], INOUT, swmod, NULL);
    tmpsw->clk2term = INPcrtTerm(clk2net, chanlist[CLKCH], CLOCK, swmod, NULL);
    tmpsw->amp2term = INPcrtTerm(amp2net, chanlist[TOP], INOUT, swmod, NULL);
    
    /* in the master of SWIN and SWOUT2, grounded terminal is always amp2term*/
    if ( libptr == NULL && amp1net->type == GROUND ) {
	/* swap amp1term and amp2term */
	tmpterm = tmpsw->amp2term;
	tmpsw->amp2term = tmpsw->amp1term;
	tmpsw->amp1term = tmpterm;
	
	/* swap clk1term and ck2term */
	tmpterm = tmpsw->clk2term;
	tmpsw->clk2term = tmpsw->clk1term;
	tmpsw->clk1term = tmpterm;
    }
    
    
    if ( tmpsw->capterm->netptr->type != GROUND ) {
	tmpsw->capterm->netptr->type = SWITCHED;
    }
    tmpsw->clk1term->netptr->type = CLOCK;
    tmpsw->clk2term->netptr->type = CLOCK;
    
    if ( libptr != NULL ) {
	/*** add the module to the list of switches of the same kind */
	INPaddLib(swmod, libptr);
    }
    
    /* add the switch to the list of switches */
    INPaddList(swmod, SWCH);
}

  INPcrtCap(name, netptr1, netptr2, value)
char *name;
struct _net *netptr1, *netptr2;
double value;
{
    struct _cap *tmpcap;
    
    Numcap++;
    
    tmpcap = ALLOC( struct _cap, 1 );
    tmpcap->cblk = NULL;
    strcpy(tmpcap->name , name);
    tmpcap->type = NOT_ASSIGNED;
    tmpcap->size = value;
    tmpcap->ratio = NOT_ASSIGNED;
    tmpcap->cratio = NOT_ASSIGNED;
    
    /*** at this point the topplate and botplate assignment is arbitrary */
    tmpcap->topplate = INPcrtTerm(netptr1, NULL, INOUT, NULL, tmpcap);
    tmpcap->botplate = INPcrtTerm(netptr2, NULL, INOUT, NULL, tmpcap);
    
    tmpcap->nuc_placed = NO;
    
    tmpcap->next = NULL;
    
    /*** no explicit capacitor list is required */
    /* the capacitors are accesed through the nets */
}

/* create a terminal and add to the list of overall input or output terminals */
  INPcrtIOterm(name, netptr, ioflag)
char *name;
struct _net *netptr;
int ioflag; /* has the value INMOD or OUTMOD */
{
    struct _module *tmpmod;
    struct _terminal *INPcrtTerm(), *tmpterm;
    int termtype;
    
    if ( ioflag == INMOD ) {
	termtype = INPUT;
    }
    else if ( ioflag == OUTMOD ) {
	termtype = OUTPUT;
    }
    else {
	printf("INPcrtIOterm;1: internal error\n");
	exit(-1);
    }
    
    tmpmod = modlist[ioflag];
    tmpterm = INPcrtTerm(netptr, NULL, termtype, tmpmod, NULL);  
    strcpy( tmpterm->name, name );
}


/* create termianl for a given module and add to the list of terminals */
struct _terminal *INPcrtTerm(netptr, channel, type, module, capacitor)
struct _net *netptr;
struct _channel *channel;
int type; /* INPUT OUTPUT INOUT */
struct _module *module;
struct _cap *capacitor;
{
    struct _terminal *tmpterm;
    struct _terminal *prevterm;
	
    Numterm++;
    
    tmpterm = ALLOC( struct _terminal, 1 );
    tmpterm->module =  module;
    tmpterm->capacitor = capacitor;
    
    /*** point to the net and add the terminal to the list */
    tmpterm->netptr = netptr;
    if ( tmpterm->netptr != NULL ) {
	tmpterm->nxtnett = netptr->firstnett;
	netptr->firstnett = tmpterm;
    }
    
    tmpterm->channel = channel;
    tmpterm->nxtcht = NULL;
    
    
    tmpterm->layer = NULL;
    tmpterm->type = type;
    tmpterm->xcoord = tmpterm->ycoord = NOT_ASSIGNED;
    tmpterm->chrlayer = NOT_ASSIGNED;
    tmpterm->chloc = NOT_ASSIGNED;
    tmpterm->chxcoord = tmpterm->chycoord = NOT_ASSIGNED;
    
    /*** the terminal is added to the end of the module terminal list */
    tmpterm->next = NULL;
    if ( module != NULL ) {
	if ( module->terminals == NULL ) {
	    module->terminals = tmpterm;
	}
	else {
	    prevterm = module->terminals;
	    while ( prevterm->next != NULL ) {
		prevterm = prevterm->next;
	    }
	    prevterm->next = tmpterm;
	}
    }
    
    return( tmpterm );
}


/* Initialize the geometric information about the modules */
INPinitMod(module) 
struct _module *module;
{
    module->row = module->col = NOT_ASSIGNED;
    module->xdim = module->ydim = NOT_ASSIGNED;
    module->xtrans = module->ytrans = NOT_ASSIGNED;
    module->transform = NO;
    module->xsep      = NOT_ASSIGNED;
    
    module->terminals = NULL;
    
    module->libptr = NULL;
}

INPinitChan( channel, chname ) 
struct _channel *channel;
int chname; 					/* TOP , BOTTOM, or CLKCH */
{
    int counter;
    
    CALLOC_ERRCHK(channel->module, 1, struct _module);
    INPinitMod( channel->module );
    
    channel->module->xtrans = 0;			/* redundant statement-hcc */
    
    channel->name = chname;
    
    channel->maxdensity = NOT_ASSIGNED;
    channel->lowchx = channel->lowchy = channel->hichx = channel->hichy =
      NOT_ASSIGNED;
    channel->numnets = NOT_ASSIGNED;
    
    /*
     * ERROR in original program
     * for ( counter = 0; counter <= MAXNUMDIR; counter++ )
     * numterms[] and tlist[] are only indexex from 0..MAXNUMDIR-1
     * In this initialization routine, since everything after it
     * is set to 0 in the structure, it probably won't make a difference. 
     * -hcc
     */
    
    for ( counter = 0; counter < MAXNUMDIR; counter++ ) {
	channel->numterms[counter] = NOT_ASSIGNED;
	channel->tlist[counter] = NULL;
    }
    
    channel->xgrids = NULL;
    channel->numvias = NOT_ASSIGNED;
    channel->vialist = NULL;
    
    channel->numwires = NOT_ASSIGNED;
    channel->snetlist = NULL;
    
    channel->ch_matrix = NULL;
    
    channel->numshpins = NOT_ASSIGNED;
    channel->shpins = NULL;
    
    channel->numshgr = NOT_ASSIGNED;
    channel->isolated_grs = NULL;
}

/* add a module to the end of its corresponding list */
INPaddList(module, listname)
struct _module *module;
int listname; /* OPAMP, CBLK or SWCH */
{
    struct _module *prevmod;
    
    if ( modlist[listname] == NULL ) {
	modlist[listname] = module;
    } else {
	prevmod = modlist[listname];
	while ( prevmod->next != NULL ) {
	    if((listname == OPAMP) &&
	       (! strcmp(prevmod->info.opamp->name, module->info.opamp->name) )) {
		errRaise("ADORE", -1, "Two opamps have the same name %s",
			 module->info.opamp->name);
	    }
	    if((listname == SWCH) &&
	       (! strcmp(prevmod->info.swch->name, module->info.swch->name) )) {
		errRaise("ADORE", -1, "Two switches have the same name %s",
			 module->info.swch->name);
	    }
	    prevmod = prevmod->next;
	}
	prevmod->next = module;
    }
}

/* create a new net and increment the global variable Numnet by 1 */
INPcrtNet(pnewnet, name)
struct _net **pnewnet;
char *name;
{
    struct _net *tmpnet;
    
    tmpnet = ALLOC( struct _net, 1);
    strcpy(tmpnet->name , name);
    /* increment the Numnet and assign it to the new net */
    tmpnet->number = ++Numnet;
    tmpnet->type = NOT_ASSIGNED;
    tmpnet->module = NULL;
    
    tmpnet->firstnett = NULL;
    /* initially all nets are not sensitive */
    tmpnet->sensitive = NO;
    
    /*** add to the beginning of the list of nets */
    tmpnet->next = netlist;
    netlist = tmpnet;
    
    *pnewnet = tmpnet;
}

/* allocate memory for a library entry */
INPcrtLibelm(pnewlib, libname)
struct _lib **pnewlib;
char *libname;
{
    struct _lib *tmplib;
    
    tmplib = ALLOC( struct _lib, 1 );
    strcpy(tmplib->libname, libname);
    
    tmplib->stptr   = NULL;
    tmplib->buslist = NULL;
  
    tmplib->next = liblist; /* Add to the beginning of library entry list */
    liblist = tmplib;
    
    *pnewlib = tmplib;
}

/* add the module to the list of modules of its library entry */
INPaddLib(module, libptr)
struct _module *module;
struct _lib *libptr;
{
    struct _mod_stack *tmpmod_st;
    
    tmpmod_st = libptr->stptr;
    libptr->stptr = ALLOC( struct _mod_stack, 1 );
    
    libptr->stptr->module = module;
    module->libptr = libptr;
    libptr->stptr->next = tmpmod_st;
    
    module->name = util_strsav( libptr->libname );
}
