/* merge the memories allocated for each set of top plates belonging
to the same cblk with the same clock phase into single memory location */
/* Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli */
/* Last Change: 8/4/87 */

/* defined: INPmergeAllTP, INPmergeTP */

#include <stdio.h>
#include "ADglobals.h"
#include "utility.h"

INPmergeAllTP()
{
	struct _module *tmpmod;
	struct _cblk *tmpcblk;


	/*** at this point the topplate and bottom plate terminals are not in the
			list of their module */

	for ( tmpmod = modlist[CBLK]; tmpmod != NULL; tmpmod = tmpmod->next ) {
		tmpcblk = tmpmod->info.cblk;

		INPmergeTP( tmpcblk->contcaps );
		INPmergeTP( tmpcblk->swcaps );
	}
}


INPmergeTP ( caplist )
struct _cap *caplist;
{
	struct _terminal *tmpterm, *prevterm;
	struct _term_stack *termstk, *tmptstk;
	struct _cap *tmpcap;
	struct _net *tmpnet;


	termstk = NULL;
	for (tmpcap= caplist; tmpcap != NULL; tmpcap = tmpcap->next) {

		/*** add the bottom plates to list for their modules */
		INPmodAddT( tmpcap->botplate );

		for ( tmptstk = termstk; tmptstk != NULL; tmptstk = tmptstk->next){
			if ( tmptstk->terminal->netptr == tmpcap->topplate->netptr ) {
				break;
			}
		}
		if ( tmptstk == NULL ) {
			CALLOC(tmptstk, 1, struct _term_stack);
			tmptstk->terminal = tmpcap->topplate;
			/* add to the stack of distinct topplate terminals */
			tmptstk->next = termstk;
			termstk = tmptstk;
		}
		else {
			/* remove topplate from its net */
			tmpnet = tmpcap->topplate->netptr;
			prevterm = tmpterm = tmpnet->firstnett;
			for ( ; tmpterm != NULL; tmpterm = tmpterm->nxtnett ) {
				if ( tmpterm == tmpcap->topplate ) {
					if ( tmpterm == tmpnet->firstnett ) {	
						tmpnet->firstnett = tmpterm->nxtnett;
					}
					else {
						prevterm->nxtnett = tmpterm->nxtnett;
					}
					break;
				}
				prevterm = tmpterm;
			}
			/*
			FREE ( tmpcap->topplate );
			*/
			tmpcap->topplate = tmptstk->terminal;
		}
	}

	/*** add the top plates to list for their modules */
	for ( tmptstk = termstk; tmptstk != NULL; tmptstk = tmptstk->next){
		INPmodAddT( tmptstk->terminal );
	}
}

/* add to the list for its module */
INPmodAddT( terminal )
struct _terminal *terminal;
{
	terminal->next = terminal->module->terminals;
	terminal->module->terminals = terminal;
}
