/* library management for automatic switched-capacitor circuit generation */
/* Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli */
/* Last change: 7/7/87 */

/*
 * Modifications to Adore
 * Henry Chang (henryc@ic.Berkeley.EDU)
 * Fall, 1990
 */

#define C_FILE_NAME "LIBfillAmp.c"

#include "ADglobals.h"
#include <stdio.h>
#include "oct.h"
#include "oh.h"

/* evaluates the dimension of the opamps and the location of
input and output ports, It also evaluates the location and width 
of the busses */

/* defined: LIBfillAmp, LIBgetBusses, LIBgetBB, LIBgetPowerTerms */


/*
 * Written by Henry Chang
 * Gets the bounding box of facet and puts that information into _lib
 * structure.
 * 19-January-91
 */
LIBgetBB (lib, facet, box)
struct _lib 	*lib;
octObject 	*facet;
octBox  	*box;
{
    int xdim,ydim;
    struct _mod_stack *mod;

    OH_ASSERT (ohBB(facet,box));
    xdim = box->upperRight.x - box->lowerLeft.x;
    ydim = box->upperRight.y - box->lowerLeft.y;
    /*
     * Put the information in modules
     */
    for (mod=lib->stptr; mod!=NULL; mod = mod->next) {
        mod->module->xdim = xdim;
        mod->module->ydim = ydim;
    } /* for */
    DEBUG (printf ("xdim = %d, ydim = %d\n", xdim, ydim));

}

/*
 * New LIBfillAmp (libptr)
 * created by Henry Chang
 * 10-December-1990
 */

LIBfillAmp (libptr)
struct _lib *libptr;
{
    octObject		facet;			/* Interface facet */
    octObject		term;			/* Formal Terminal */
    octGenerator	gen;			/* Generator for Terminals */
    octBox		box;			/* Terminal Box */
    octStatus		gen_return;		/* Geneartor return status */
    octStatus		ohBB_return;		/* ohBB return status */
    int			inplus_flg;
    int			inminus_flg;
    int			outplus_flg;
    int			outminus_flg;	
    int			x,y;			/* x, y coordinates */

    struct _mod_stack	*mod;			/* current module */
    struct _opamp	*opamp;			/* current opamp */

    octObject		contents_facet;		/* for LIBgetBusses */

    octBox              facet_bb;               /* facet bounding box */

    octBegin ();

    ohUnpackDefaults (&facet,"r",":physical:interface");
    ohUnpackFacetName (&facet, libptr->libname); 
    facet.contents.facet.facet = "interface";
    VERBOSE (printf ("Processing the library entry, %s:%s:%s.\n", 
	facet.contents.facet.cell, facet.contents.facet.view,
	facet.contents.facet.facet));
    OH_ASSERT (octOpenFacet (&facet));

    /* 
     * Get Bounding Box of the OpAmp 
     */

    LIBgetBB (libptr,&facet,&facet_bb);

    /* 
     * Get the Terminal Locations of the OpAmp
     * Expects terminals on the top of the cell and on a specific layer
     */

    OH_ASSERT (octInitGenContents(&facet,OCT_TERM_MASK,&gen));
    inplus_flg = inminus_flg = outplus_flg = outminus_flg = 0; 
    while ( (gen_return=octGenerate (&gen, &term)) == OCT_OK) {
        ohBB_return = ohBB (&term, &box);
	if (ohBB_return != OCT_OK) {
           if (ohBB_return == OCT_NO_BB)
              errRaise (C_FILE_NAME,ohBB_return, "No Bounding Box.");
           errRaise (C_FILE_NAME,ohBB_return, "ohBB failed.");
        }
        x = (box.lowerLeft.x + box.upperRight.x ) / 2;
        y = (box.lowerLeft.y + box.upperRight.y ) / 2;
        /*
         * Original Program had:
         * y = box.upperRight.y
         * -hcc
         */
        if ( strcmp(term.contents.term.name,"INPUT_PLUS")==0 ) {
           inplus_flg++;
           DEBUG (printf ("in_plus=(%d,%d)\n",x,y));
           for (mod=libptr->stptr; mod!=NULL; mod = mod->next) {
              opamp = mod->module->info.opamp;
              opamp->inp[PLUS]->xcoord = x;
              opamp->inp[PLUS]->ycoord = y;
              opamp->inp[PLUS]->chrlayer = 1;
              opamp->inp[PLUS]->chloc = BOTTOM; 
              strcpy(opamp->inp[PLUS]->name,term.contents.term.name);
           } /* for */ 
	} /* INPUT_PLUS */
        else if ( strcmp(term.contents.term.name,"INPUT_MINUS")==0 ) {
           inminus_flg++;
           DEBUG (printf ("in_minus=(%d,%d)\n",x,y));
           for (mod=libptr->stptr; mod!=NULL; mod = mod->next) {
              opamp = mod->module->info.opamp;
              opamp->inp[MINUS]->xcoord = x;
              opamp->inp[MINUS]->ycoord = y;
              opamp->inp[MINUS]->chrlayer = 1;
              opamp->inp[MINUS]->chloc = BOTTOM; 
              strcpy(opamp->inp[MINUS]->name,term.contents.term.name);
           } /* for */ 
	} /* INPUT_MINUS */
        else if ( strcmp(term.contents.term.name,"OUTPUT_PLUS")==0 ) {
           outplus_flg++;
           DEBUG (printf ("out_plus=(%d,%d)\n",x,y));
           for (mod=libptr->stptr; mod!=NULL; mod = mod->next) {
              opamp = mod->module->info.opamp;
              opamp->out[PLUS]->xcoord = x;
              opamp->out[PLUS]->ycoord = y;
              opamp->out[PLUS]->chrlayer = 1;
              opamp->out[PLUS]->chloc = BOTTOM; 
              strcpy(opamp->out[PLUS]->name,term.contents.term.name);
           } /* for */ 
	} /* OUTPUT_PLUS */
        else if ( strcmp(term.contents.term.name,"OUTPUT_MINUS")==0 ) {
           outminus_flg++;
           DEBUG (printf ("out_minus=(%d,%d)\n",x,y));
           for (mod=libptr->stptr; mod!=NULL; mod = mod->next) {
              opamp = mod->module->info.opamp;
              opamp->out[MINUS]->xcoord = x;
              opamp->out[MINUS]->ycoord = y;
              opamp->out[MINUS]->chrlayer = 1;
              opamp->out[MINUS]->chloc = BOTTOM; 
              strcpy(opamp->out[MINUS]->name,term.contents.term.name);
           } /* for */ 
	} /* OUTPUT_MINUS */
        else {
	   VERBOSE (printf("Non-signal Terminal, %s, found at (%d,%d).\n",
              term.contents.term.name, x, y));    
        } /* else */
    } /* while */ 
    octFreeGenerator (&gen);
    if (gen_return != OCT_GEN_DONE) {
        errRaise (C_FILE_NAME,0,"Did not receive OCT_GEN_DONE");
    }

   /*
    * Check to make sure the four terminals that are required were found.
    */

    if (inplus_flg==0) {
        errRaise 
	   (C_FILE_NAME,0,"Did not find input(+) terminal on opamp, %s.",
           libptr->libname); 
    }        
    if (inminus_flg==0) {
        errRaise 
	   (C_FILE_NAME,0,"Did not find input(-) terminal on opamp, %s.",
           libptr->libname); 
    }        
    if (outplus_flg==0) {
        errRaise 
	   (C_FILE_NAME,0,"Did not find output(+) terminal on opamp, %s.",
           libptr->libname); 
    }        
    if (outminus_flg==0) {
       /*
        * errRaise 
	*   (C_FILE_NAME,0,"Did not find output(-) terminal on opamp, %s.",
        *   libptr->libname); 
        */
        VERBOSE (printf 
           ("Warning: Did not find output(-) terminal on opamp, %s.\n",
              libptr->libname));
    }       
    if ( (inplus_flg > 1) || (inminus_flg > 1) || (outplus_flg > 1) ||
         (outminus_flg > 1) ) {
	errRaise
	   (C_FILE_NAME,0,"Duplicate Formal Terminal found on opamp, %s.",
           libptr->libname);
    } 

   /*
    * Get Busses Bag
    * Searh for SUPPLY and GROUND nets
    * Have to look in both the contents facet and the interface facet.
    */

    ohUnpackDefaults (&contents_facet,"r",":physical:contents");
    ohUnpackFacetName (&contents_facet, libptr->libname);    
    contents_facet.contents.facet.facet = "contents";
    VERBOSE (printf ("Processing the library entry, %s:%s:%s.\n", 
	contents_facet.contents.facet.cell, facet.contents.facet.view,
	contents_facet.contents.facet.facet));
    OH_ASSERT (octOpenFacet (&contents_facet));

    LIBgetPowerTerms (libptr,&contents_facet,&facet,&facet_bb);
    LIBgetBusses(&contents_facet, &facet, libptr);

    OH_ASSERT (octCloseFacet (&contents_facet));
    OH_ASSERT (octCloseFacet (&facet));
    octEnd();

} /* LIBfillAmp */

/* process the busses for a given library element */
LIBgetBusses(olibobcont, olibobject, libelm)
octObject *olibobcont, *olibobject;
struct _lib *libelm; /* the bus belongs to this library element */
{
  octObject obag;
  octObject obox;
  octObject olayer;
  octObject odummy;
  octStatus status;
  octGenerator ogenbox, ogenerator;


  obag.type = OCT_BAG;
  obag.contents.bag.name = "BUSSES";
  if ( octGetByName(olibobject, &obag) < OCT_OK ) {
     /*		
      * octError("getting the BUSSES bag");
      * exit(-1);  
      * -enrico
      */
    if(olibobcont == NULL) {
       VERBOSE(printf("No BUSSES bag found\n"));
    }
    else { 
       LIBgetBusses(NULL, olibobcont, libelm);
    }
    return;
  }

  if ( octInitGenContents(&obag,OCT_BOX_MASK,&ogenbox) < OCT_OK ) {
    octError("generating box from BUSSES bag");
    exit(-1);
  }

  while (( status = octGenerate(&ogenbox, &obox)) == OCT_OK ) {
    
    if (octInitGenContainers(&obox,OCT_LAYER_MASK,&ogenerator) 
	< OCT_OK ) {
      octError("generating layer from a box");
      exit(-1);
    }

    if ( octGenerate(&ogenerator,&olayer) < OCT_OK ) {
      octError("layer generation from a box");
      exit(-1);
    }

    if ( octGenerate(&ogenerator,&odummy) != OCT_GEN_DONE ) {
      octError("more than one layer associated with a box");
      exit(-1);
    }

    LIBaddBus(olayer.contents.layer.name, obox.contents.box.lowerLeft.y,
	      obox.contents.box.upperRight.y,libelm);
  }
  
  if ( status != OCT_GEN_DONE ) {
    octError("box generation from BUSSES bag");
    exit(-1);
  }
}

/* add a bus to the list of busses for it library */
LIBaddBus( layer, lowlefty, uprighty, libelm)
char *layer;
int lowlefty, uprighty;
struct _lib *libelm;
{
	struct _bus *tmpbus;

	CALLOC( tmpbus, 1, struct _bus);
	strcpy(tmpbus->layer, layer);
	tmpbus->lowlefty = lowlefty;
	tmpbus->uprighty = uprighty;

	tmpbus->next = libelm->buslist;
	libelm->buslist = tmpbus;
}

/*
 * Written by Henry Chang
 * Gets the SUPPLY and GROUND nets. 
 * Store into _lib structure.
 * 19-January-91
 */
LIBgetPowerTerms (lib,contents,interface,facet_bb)
struct _lib	*lib;
octObject	*contents;
octObject	*interface;
octBox		*facet_bb;
{
    octGenerator        gen;                    /* Generator for Terminals */
    octStatus           gen_return;             /* Geneartor return status */

    struct _mod_stack   *mod;                   /* current module */

    struct _sterminal   *first_sterm;           /* Supply Terminals */
    struct _sterminal   *sterm, *prev_sterm;

    octObject           snet;                   /* Supply net */
    octObject           sprop;                  /* Supply Property */
    octObject           sfterm;                 /* Supply Formal terminals */
    octGenerator        sfterm_gen;             /* sfterm generator */
    octStatus           sfterm_status;          /* generator status */
    int                 sside;                  /* Supply term location */
 
    /*
     * Search for SUPPLY and GROUND terminals
     */
    first_sterm = NULL;
    OH_ASSERT (octInitGenContents(contents,OCT_NET_MASK,&gen));
    while ( (gen_return=octGenerate (&gen, &snet)) == OCT_OK) {
        if (strcmp(snet.contents.net.name,"") != 0) {
            if (ohGetByPropName(&snet,&sprop,"NETTYPE")==OCT_OK) {
                if ( (strcmp(sprop.contents.prop.value.string,"SUPPLY")==0) ||
                     (strcmp(sprop.contents.prop.value.string,"GROUND")==0) ) {
                    OH_ASSERT (octInitGenContents(&snet,OCT_FORMAL_MASK,
                               &sfterm_gen));
                    while ( (sfterm_status=
                             octGenerate (&sfterm_gen, &sfterm)) == OCT_OK) {
                       prev_sterm = sterm;
                       CALLOC_ERRCHK (sterm,1,struct _sterminal);
                       if (first_sterm==NULL) {
                          first_sterm = sterm;
                       }
                       else {
                          prev_sterm->next = sterm;
                       }
                       LIBgetFtermSide (interface,&sfterm,facet_bb,&sside);
                       /* Note that this is passing in the interface facet */
                       strcpy (sterm->name,sfterm.contents.term.name);
                       strcpy (sterm->net,snet.contents.net.name);
                       strcpy (sterm->property,
                               sprop.contents.prop.value.string);
                       sterm->side = sside;
                       sterm->next = NULL;
                       DEBUG (printf (
                          "net: %s, NETTYPE: %s, term: %s, side: %d\n",
                          sterm->net, sterm->property, sterm->name,
                          sterm->side));
                    } /* while terminals */
                    octFreeGenerator (&sfterm_gen);
                    if (sfterm_status != OCT_GEN_DONE) {
                       errRaise (C_FILE_NAME,0,"Did not receive OCT_GEN_DONE");
                    }
                } /* SUPPLY or GROUND found */
            } /* Property found */
        } /* if */
    } /* while */ 
    octFreeGenerator (&gen);
    if (gen_return != OCT_GEN_DONE) {
        errRaise (C_FILE_NAME,0,"Did not receive OCT_GEN_DONE");
    }
    /*
     * Put the information in modules
     */
    for (mod=lib->stptr; mod!=NULL; mod = mod->next) {
        mod->module->sterminals = first_sterm;
    } /* for */ 
} /* LIBgetPowerTerms */

/*
 * LIBgetFtermSide
 * This routine reads the bounding box of facet
 * It reads the bouding box of fterm, and then tries to determine which
 * side the terminals are on, and returns its side.
 * It will preferentially choose the left or the right side if terminals
 * are found on the corner.
 * This routine is unpredictable if terminal edge is equally close to 3 or more
 * bounding box edges.
 */
LIBgetFtermSide (facet,fterm,facet_bb,side)
octObject *facet;
octObject *fterm;
octBox *facet_bb;
int *side;
{
   struct _sides {
       int	d;				/* Distance to Edge */
       int	side;				/* Side             */	 
   };

   octBox 		fterm_bb;
   octObject		term;
   octStatus		ohBB_return;
   struct _sides	*list[4];
   struct _sides	*tmp_list;
   int			i,j;

   DEBUG (printf ("Terminal: %s   ",fterm->contents.term.name));
   OH_ASSERT(ohGetByTermName (facet,&term,fterm->contents.term.name));
   ohBB_return = ohBB(&term,&fterm_bb);
   if (ohBB_return != OCT_OK) {
       if (ohBB_return == OCT_NO_BB)
           errRaise (C_FILE_NAME,ohBB_return, "No Bounding Box.");
       errRaise (C_FILE_NAME,ohBB_return, "ohBB failed.");
   }
   /* ohBB_return = ohBB(fterm,&fterm_bb); Don't know why this won't work. */
   DEBUG (printf ("facet: (%d,%d)-(%d,%d)   fterm: (%d,%d)-(%d,%d)\n",
      facet_bb->lowerLeft.x, facet_bb->lowerLeft.y,
      facet_bb->upperRight.x, facet_bb->upperRight.y,
      fterm_bb.lowerLeft.x, fterm_bb.lowerLeft.y,
      fterm_bb.upperRight.x, fterm_bb.upperRight.y));
   for (i=0; i<4; i++) {
      CALLOC_ERRCHK (list[i], 1, struct _sides);
   }
   list[0]->d = fterm_bb.lowerLeft.x - facet_bb->lowerLeft.x;
   list[0]->side = LEFT;
   list[1]->d = facet_bb->upperRight.x - fterm_bb.upperRight.x;
   list[1]->side = RIGHT;
   list[2]->d = facet_bb->upperRight.y - fterm_bb.upperRight.y;
   list[2]->side = TOP;
   list[3]->d = facet_bb->upperRight.y - fterm_bb.upperRight.y;
   list[3]->side = BOTTOM;
   /* sort */
   for (i=0; i<3; i++) {
      for (j=i; j<4; j++) {
         if (list[j]->d < list[i]->d) {
            tmp_list = list[i];
            list[i] = list[j];
            list[j] = tmp_list;
         } /* if */
      } /* for j */
   } /* for i */
   *side = list[0]->side;
   /* Check for preferential right or left */
   if (list[0]->d == list[1]->d) {
      if ((list[0]->side==RIGHT) || (list[1]->side==RIGHT)) {
         *side = RIGHT;
      } 
      if ((list[0]->side==LEFT) || (list[1]->side==LEFT)) {
         *side = LEFT;
      } 
   } /* Have to pick preferential side */  

} /* LIBgetFtermSide */
