/* find the optimal permutation of capacitor blocks to minimize
the total estimated net length */
/* Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincetelli */
/* Last Change: 8/5/87 */

/* defined: PERMcapNets, PERMmarkMatrix */

#include "PERMglobals.h"
#include <stdio.h>


/* each entry is on if the corresponding cblk belongs to the net */
set_t **net_matrix;

int numnewnets; /* each net corresponds to one switch or one opamp */


/* find which blocks are directly or through an opamp or a switch 
connected to other blocks */

PERMcapNets() 
{
    int newnet;
    int counter;
    int direct_con; /* signifies direct connecton of input to cblk */

    char *set_format_bv();
    struct _module *tmpmod;
    struct _net *tmpnet;
    struct _terminal *tmpterm;
    set_t *set_new();


    /*** initialize the newnet matrix */
    /*** there is one net per opamp and one net per capacitor */
    /*** there might be an extra net for dircet input to cblk connections */
    numnewnets = Numinteg + Numsw + 1;
    CALLOC(net_matrix, Numcblk + 2, set_t *);
    for ( counter = 0; counter <= Numcblk + 1; counter++) {
        net_matrix[counter] = set_new(numnewnets);
    }


    newnet = 1;
    tmpmod = modlist[OPAMP];
    for ( ; tmpmod != NULL; newnet++, tmpmod = tmpmod->next) {
        tmpterm = tmpmod->terminals;
        for ( ; tmpterm != NULL; tmpterm = tmpterm->next ) {
            tmpnet = tmpterm->netptr;
            if ( tmpnet->type != GROUND ) {
                if (PERMmarkMatrix(tmpnet, newnet, chanlist[BOTTOM]) == NO ) {
                    printf("PERMcapNets: isolated opamp terminal\n");
                    gerror = ON;
                }
            }
        }
    }
    tmpmod = modlist[SWCH];
    for ( ; tmpmod != NULL; newnet++, tmpmod = tmpmod->next) {
        tmpterm = tmpmod->terminals;
        for ( ; tmpterm != NULL; tmpterm = tmpterm->next ) {
            tmpnet = tmpterm->netptr;
            if ( tmpnet->type != GROUND && tmpnet->type != CLOCK ) {
                if (PERMmarkMatrix(tmpnet, newnet, chanlist[TOP]) == NO ) {
                    printf("PERMcapNets: isolated switch terminal\n");
                    gerror = ON;
                }
            }
        }
    }

    /*** there is no direct connection of capacitor blocks */
    /*** there might be direct connections from INMOD to some cblk */
    /*** the overall input cannot directly connect to the opamp's input */
    newnet = 0;
    tmpterm = modlist[INMOD]->terminals;
    direct_con = NO;
    for ( ; tmpterm != NULL; tmpterm = tmpterm->next ) {
        tmpnet = tmpterm->netptr;
        if ( tmpnet->type != GROUND ) {
            direct_con = PERMmarkMatrix(tmpnet, newnet, BOTTOM);
        }
    }
}

/* for each block belonging to orignet mark the entry belonging to newnet */
PERMmarkMatrix(orignet, newnet, channel)
struct _net *orignet;
int newnet;
struct _channel *channel; /* the switch or opamp channel */
{
    struct _terminal *tmpterm;
    int cblk_connection; /* a given net has capacitor terminals or not */

    cblk_connection = NO;
    tmpterm = orignet->firstnett;
    for ( ; tmpterm != NULL; tmpterm = tmpterm->nxtnett ) {
        if ( tmpterm->channel == channel ) {
            if ( tmpterm->module->type == CBLK ) {
                set_insert(net_matrix[tmpterm->module->col], newnet);
                cblk_connection = YES;
            }
            else if ( tmpterm->module->type == INMOD ) {
                set_insert(net_matrix[0], newnet);
                cblk_connection = YES;
            }
            else if ( tmpterm->module->type == OUTMOD ) {
                set_insert(net_matrix[Numcblk + 1], newnet);
                cblk_connection = YES;
            }
        }
    }
    return ( cblk_connection );
}
