/* assign columns for a channel, for adore2 */
/* Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli */
/* Last Change: 7/30/87 */

#include "ADglobals.h"
#include "TEKtech.h"
#include <stdio.h>

int debug;

ROUTassCol( channel )
struct _channel *channel;
{
	int counter;
	int topx, botx, PERMfindXcrd();
	int numgrids;
	int maxnumgrids;
	int gridx, termx;
	int grid_dist;

	struct _terminal *tmptopt, *tmpbott;
	struct _terminal *tmpterm;
	struct _term_stack *chterms, *tmpcht;
	struct _xgrids *gridlist;

	debug = NO;

	/* adjust the position of the pins by evaluating value of xoffset for
	each pin so that all the addjusted possitions are at least minctctsp apart*/
	ROUTfindOffset( channel->tlist[TOP] );
	ROUTfindOffset( channel->tlist[BOTTOM] );


	/*** generate a sorted list of top and bottom channel terminals */

	tmptopt = channel->tlist[TOP];
	tmpbott = channel->tlist[BOTTOM];
	CALLOC(chterms, 1, struct _term_stack);
	tmpcht = chterms;
	while ( tmptopt != NULL && tmpbott != NULL ) {
		topx = PERMfindXcrd(tmptopt) + tmptopt->xoffset;
		botx = PERMfindXcrd(tmpbott) + tmpbott->xoffset;
		if ( topx < botx ) {
			tmpcht->terminal = tmptopt;
			tmptopt = tmptopt->nxtcht;
		}
		else if ( topx > botx ) { 
			tmpcht->terminal = tmpbott;
			tmpbott = tmpbott->nxtcht;
		}
		else {
			/* they have same coordinates, top terminal comes first */
			tmpcht->terminal = tmptopt;
			tmptopt = tmptopt->nxtcht;
		}
		CALLOC(tmpcht->next, 1, struct _term_stack);
		tmpcht = tmpcht->next;
	}
	if ( tmptopt == NULL && tmpbott != NULL ) {
		while ( tmpbott != NULL ) {
			tmpcht->terminal = tmpbott;
			tmpbott = tmpbott->nxtcht;
			CALLOC(tmpcht->next, 1, struct _term_stack);
			tmpcht = tmpcht->next;
		}
	}
	else if ( tmptopt != NULL && tmpbott == NULL ) {
		while ( tmptopt != NULL ) {
			tmpcht->terminal = tmptopt;
			tmptopt = tmptopt->nxtcht;
			CALLOC(tmpcht->next, 1, struct _term_stack);
			tmpcht = tmpcht->next;
		}
	}
	tmpcht->terminal = NULL;
	tmpcht->next = NULL;


	/*** now generate the grid lines */
	/*** assumption is all the pins in each row are at least minctcsp apart */

	maxnumgrids = (modlist[OUTMOD]->xtrans + modlist[OUTMOD]->xsep) / minctctsp
					+ channel->numterms[TOP] + channel->numterms[BOTTOM];
	CALLOC(channel->xgrids, maxnumgrids, struct _xgrids); 

	if ( chterms->terminal == NULL ) {
		printf("ROUTassCol: no pins associated with a channel\n");
		gerror = ON;
	}
	channel->xgrids[0].xcoord = 0;
	channel->xgrids[1].xcoord = channel->xgrids[0].xcoord;
	numgrids = 1;
	for ( tmpcht = chterms; tmpcht->terminal != NULL; tmpcht = tmpcht->next ) {
		gridx = channel->xgrids[numgrids].xcoord + minctctsp;
		termx = PERMfindXcrd(tmpcht->terminal) + tmpcht->terminal->xoffset;
		while ( gridx <= termx ) {
			numgrids++;
			channel->xgrids[numgrids].xcoord = gridx;
			channel->xgrids[numgrids].topterm =
							channel->xgrids[numgrids].botterm = NULL;
			gridx += minctctsp;
		}
		if ( tmpcht->terminal->chloc == TOP ) {
			channel->xgrids[numgrids].topterm = tmpcht->terminal;
			tmpcht->terminal->chxcoord = numgrids;
			if ( channel->xgrids[numgrids].botterm == NULL ) {
				channel->xgrids[numgrids].xcoord = termx;
			}
		}
		else if ( tmpcht->terminal->chloc == BOTTOM ) {
			channel->xgrids[numgrids].botterm = tmpcht->terminal;
			tmpcht->terminal->chxcoord = numgrids;
			if ( channel->xgrids[numgrids].topterm == NULL ) {
				channel->xgrids[numgrids].xcoord = termx;
			}
		}
		else {
			fprintf(stderr,"ROUTassCol;1: internal error\n");
			exit(-1);
		}
	}
	gridx = channel->xgrids[numgrids].xcoord + minctctsp;
	while ( gridx <= modlist[OUTMOD]->xtrans ) {
		numgrids++;
		channel->xgrids[numgrids].xcoord = gridx;
		gridx += minctctsp;
	}
	channel->xgrids[numgrids + 1].xcoord = modlist[OUTMOD]->xtrans +
											modlist[OUTMOD]->xsep;

	channel->lowchx = 0;
	channel->hichx = numgrids + 1;


	/*** this is a post processing stage. For every pin, if its translated
	xcoordinate minus the coordinate of a neighboring  grid whose pin is
	unoccupied, is less than minctctsp/2, change the grid line assignment
	of the pin to the neighboring grid line */

	for ( counter = 2; counter <= (numgrids - 1); counter++) {
		if ( channel->xgrids[counter].topterm != NULL &&
					channel->xgrids[counter + 1].topterm == NULL ) {
			grid_dist = abs(PERMfindXcrd(channel->xgrids[counter].topterm) -
					channel->xgrids[counter+1].xcoord);
			if ( (2 * grid_dist) < minctctsp ) {
				channel->xgrids[counter + 1].topterm =
						channel->xgrids[counter].topterm;
				channel->xgrids[counter + 1].topterm->chxcoord = counter + 1;
				channel->xgrids[counter].topterm = NULL;
			}
		}
		if ( channel->xgrids[counter].botterm != NULL &&
					channel->xgrids[counter + 1].botterm == NULL ) {
			grid_dist = abs(PERMfindXcrd(channel->xgrids[counter].botterm) -
					channel->xgrids[counter+1].xcoord);
			if ( (2 * grid_dist) < minctctsp ) {
				channel->xgrids[counter + 1].botterm =
						channel->xgrids[counter].botterm;
				channel->xgrids[counter + 1].botterm->chxcoord = counter + 1;
				channel->xgrids[counter].botterm = NULL;
			}
		}
	}

	if ( debug == YES ) {
		printf("PERMassCol:chterms:\nnet  chloc  xcrd  chxcrd  chycrd layer\n");
		tmpcht = chterms;
		for ( ; tmpcht->terminal != NULL; tmpcht = tmpcht->next ) {
			tmpterm = tmpcht->terminal;
			termx = PERMfindXcrd(tmpterm);
			printf("PERMassCol: %d %d %d %d %d %d\n", 
								tmpterm->netptr->number, tmpterm->chloc, termx,
								tmpterm->chxcoord, tmpterm->chycoord,
								tmpterm->chrlayer);
		}
	}
}


/* adjust the position of the pins by evaluating value of xoffset for
each pin so that all the addjusted possitions are at least minctctsp apart*/

ROUTfindOffset( termlist ) 
struct _terminal *termlist;
{
	int prevxcrd, curxcrd, nxtxcrd;
	struct _terminal *tmpterm;


	if ( termlist == NULL ) {
		return;
	}

	for ( tmpterm = termlist; tmpterm != NULL; tmpterm = tmpterm->nxtcht) {
		tmpterm->xoffset = 0;
	}

	prevxcrd = 0;
	curxcrd = PERMfindXcrd(termlist) + termlist->xoffset;
	tmpterm = termlist;
	for ( ; tmpterm->nxtcht != NULL; tmpterm = tmpterm->nxtcht ) {
		nxtxcrd = PERMfindXcrd( tmpterm->nxtcht ) + tmpterm->nxtcht->xoffset;
		if ( (nxtxcrd - curxcrd) < minctctsp ) {
			if ( (nxtxcrd - prevxcrd) >= (2 * minctctsp) ) {
				tmpterm->xoffset -= minctctsp - ( nxtxcrd - curxcrd );
			}
			else {
				printf("ROUTfindOffset: pins too close\n");
				/* exit(-1); -hcc*/
				/*
				 * I've taken this out so that it will run.
				 */
			}
		}
		prevxcrd = curxcrd;
		curxcrd = nxtxcrd;
	}
}
