/* read the output of channel router for adore2 */
/* Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli */
/* Last Change: 8/9/87 */

/*
 * Modifications to Adore
 * Henry Chang (henryc@ic.Berkeley.EDU)
 * Fall, 1990
 */

#define C_FILE_NAME "ROUTchrOut.c"

#include "ADglobals.h"
#include "ROUTglobals.h"
#include "TEKtech.h"
#include "oct.h"
#include <stdio.h>
#include "oh.h"
#include "utility.h"

/*  set up the data strucure describing the channel */

ROUTchrOut(channel,tmp_dir)
struct _channel *channel; /* the channel to be layed out */
char *tmp_dir;
{
  int counter;
  int numvias, numwires;
  int net;
  int chx1, chy1, chx2, chy2;
  int chxl, chyl, chxh, chyh;
  int tmpchx, tmpchy;
  int layer, layer1, layer2;

  char newstring[40];
  struct _snet_stack *tmpsnet_stack;
  char filename[FILENAME_SIZE];
  FILE *fp, *fopen();		/* the file information is read from */

  sprintf (filename,"%s/%s", tmp_dir, CH_OUT_FILE);
  fp = fopen(filename, "r");

  fscanf(fp,"%s",newstring);
  while ( strcmp("channelwiring", newstring) != 0 ) {
    if (strcmp("failed",newstring) == 0 ) { 
      fclose ( fp );
      return( FAILURE );
    }
    if(fscanf(fp,"%s",newstring) < 1) {
      strcpy(newstring, "failed"); /*  Modified by Enrico 10/30/1990 */
    }
  }

  fscanf(fp,"%s %d",newstring, &numvias);
  if ( strcmp("vias",newstring ) != 0 ) {
    fprintf(stderr,"ROUTchrOut;1: internal error\n");
    exit( -1);
  }

  channel->numvias = numvias;
  CALLOC(channel->vialist, numvias + 1, struct _via );

  for ( counter = 1; counter <= numvias ; counter++ ) {
    /*** read coordinates of the vias center */
    fscanf(fp,"%d %d %d %d %d",&net,&chx1,&chy1,&layer1,&layer2);

    /*** the vias' channel coordinates are stored */
    channel->vialist[counter].net = net;
    channel->vialist[counter].chx = chx1;
    channel->vialist[counter].chy = chy1;
    channel->vialist[counter].layer1 = layer1;
    channel->vialist[counter].layer2 = layer2;

    channel->ch_matrix[chx1][chy1].via = channel->vialist + counter;
  }


  fscanf(fp,"%s %d",newstring, &numwires);
  if ( strcmp("wires",newstring ) != 0 ) {
    fprintf(stderr,"ROUTchrOut;2: internal error\n");
    exit( -1);
  }

  channel->numwires = numwires;
  CALLOC(channel->snetlist, numwires + 1, struct _subnet );

  for ( counter = 1; counter <= numwires ; counter++ ) {
    fscanf(fp,"%d %d %d %d %d %d",&net, &chx1, &chy1, &chx2, &chy2, &layer);

    /*** the subnet's channel information are stored */
    channel->snetlist[counter].net = net;
    chxl = channel->snetlist[counter].chxl = min(chx1, chx2);
    chyl = channel->snetlist[counter].chyl = min(chy1, chy2);
    chxh = channel->snetlist[counter].chxh = max(chx1, chx2);
    chyh = channel->snetlist[counter].chyh = max(chy1, chy2);
    channel->snetlist[counter].layer = layer;

    if ( chx1 == chx2 && chy1 != chy2 ) {
      /* vertical wire */
      channel->snetlist[counter].type = VERT;
    }
    else if ( chx1 != chx2 && chy1 == chy2 ) {
      /* horizontal wire */
      channel->snetlist[counter].type = HORIZ;
      /* find if this subnet goes outside the filter */
      ROUTioSubnet(channel, channel->snetlist + counter );
    }
    else {
      fprintf(stderr,"ROUTchrOut;3: internal error\n");
      exit(-1);
    }

    /*** add the subnet to grids that it crosses */
    for ( tmpchx = chxl; tmpchx <= chxh; tmpchx++ ) {
      for ( tmpchy = chyl; tmpchy <= chyh; tmpchy++ ) {
	CALLOC(tmpsnet_stack, 1, struct _snet_stack );
	tmpsnet_stack->subnet = channel->snetlist + counter;
	tmpsnet_stack->next = 
	  channel->ch_matrix[tmpchx][tmpchy].snet_stack;
	channel->ch_matrix[tmpchx][tmpchy].snet_stack = tmpsnet_stack;
      }
    }
  }

  fclose( fp );
  return( SUCCESS );
}


/* if the subnet is incident to the boundary, find the ycoordinate of IO pin */
ROUTioSubnet(channel, subnet )
struct _channel *channel;
struct _subnet *subnet;
{
  struct _terminal *tmpterm;


  if ( subnet->chxl == channel->lowchx || subnet->chxh == channel->hichx ) {
    tmpterm = ordered_nets[subnet->net]->firstnett;
    for ( ; tmpterm != NULL; tmpterm = tmpterm->nxtnett ) {
      if ( tmpterm->channel == channel ) {
	if ( tmpterm->module->type == INMOD || 
	    tmpterm->module->type == OUTMOD ) {
	  tmpterm->ycoord = minctctsp * subnet->chyl;
	}
      }
    }
  }
}


/* the following procedure converts the channel router's output
of the form generated by mighty to oct format */

octObject Ochannel;

ROUTgenLay( channel,instances )
struct _channel *channel;
char *instances;
{
  int lowleftx, lowlefty, uprightx, uprighty; /*coordinates of a box */
  int nlowleftx, nlowlefty, nuprightx, nuprighty; /*coordinates of a box */
  int xlcrd, xhcrd, ylcrd, yhcrd;
  int tmpnum;

  int counter;
  int width;
  char *layername;
  char tmpname[40], chname[100], *CBLKpushDir();

  struct _subnet *tmpsnet;
  struct _via *tmpvia;

  tmpnum = channel->name;
  sprintf(tmpname, "chan_%d", tmpnum);
  strcpy(chname , CBLKpushDir(instances, tmpname));
  DEBUG (printf ("Opening the Facet %s with the \"w\" option...\n",chname));
  channel->module->name = util_strsav( chname );
  openfacet( &Ochannel, chname, "physical", "contents", "w");

  for ( counter = 1; counter <= channel->numwires; counter++) {
    tmpsnet = channel->snetlist + counter;
    if ( tmpsnet->type == VERT ) {
      xlcrd = xhcrd = ROUTvertXcrd(channel, tmpsnet);
    }
    else if ( tmpsnet->type == HORIZ ) {
      xlcrd = ROUThorizXcrd(channel, tmpsnet, LEFT);
      xhcrd = ROUThorizXcrd(channel, tmpsnet, RIGHT);
    }
    ylcrd = tmpsnet->chyl * minctctsp;
    yhcrd = tmpsnet->chyh * minctctsp;

    ROUTaddSnet(channel, xlcrd, ylcrd, xhcrd, yhcrd, tmpsnet->layer);
  }

  /*** generate layout for the vias */
  for ( counter = 1; counter <= channel->numvias ; counter++ ) {
    tmpvia = channel->vialist + counter;

    /* convert channel coordinates to actual coordinates */
    xlcrd = channel->xgrids[tmpvia->chx].xcoord;
    ylcrd = tmpvia->chy * minctctsp;

    /* for now assume via has the same dimensions as the contact */

    lowleftx = xlcrd - mincontd / 2;
    lowlefty = ylcrd - mincontd / 2;
    uprightx = lowleftx + mincontd;
    uprighty = lowlefty + mincontd;
    makebox(&Ochannel, con_L, lowleftx, lowlefty, uprightx, uprighty);

    nlowleftx = lowleftx - metovcont;
    nlowlefty = lowlefty - metovcont;
    nuprightx = uprightx + metovcont;
    nuprighty = uprighty + metovcont;
    makebox(&Ochannel, metal_L, nlowleftx, nlowlefty, nuprightx, nuprighty);

    nlowleftx = lowleftx - polyovcon;
    nlowlefty = lowlefty - polyovcon;
    nuprightx = uprightx + polyovcon;
    nuprighty = uprighty + polyovcon;
    makebox(&Ochannel, botpoly_L,nlowleftx,nlowlefty, nuprightx, nuprighty);
  }

  OH_ASSERT(octCloseFacet(&Ochannel));
  /* GenInterface (&Ochannel); */
}

ROUTvertXcrd(channel, subnet)
struct _channel *channel;
struct _subnet *subnet;
{
  int topnet, botnet;
  int tmpxcrd;
  int xlcrd, ylcrd, xhcrd, yhcrd;

  struct _xgrids *tmpxgrid;
  struct _via *tmpvia;


  tmpxgrid = channel->xgrids + (subnet->chxl);

  /* net numbers are always positive */
  topnet = botnet = -1;
  if ( tmpxgrid->topterm != NULL ) {
    topnet = tmpxgrid->topterm->netptr->number;
  }
  if ( tmpxgrid->botterm != NULL ) {
    botnet = tmpxgrid->botterm->netptr->number;
  }

  if ( subnet->chyh == (channel->hichy - 1) && subnet->net == topnet ) {
    tmpxcrd = PERMfindXcrd(tmpxgrid->topterm);
    tmpvia = channel->ch_matrix[subnet->chxh][channel->hichy - 1].via;
    if ( tmpvia == NULL && tmpxcrd != tmpxgrid->xcoord ) {
      /* make the extra horizontal subnet */
      xlcrd = min(tmpxcrd, tmpxgrid->xcoord);
      xhcrd = max(tmpxcrd, tmpxgrid->xcoord);
      ylcrd = yhcrd = (channel->hichy - 1) * minctctsp;
      ROUTaddSnet(channel, xlcrd, ylcrd, xhcrd, yhcrd, 
		  tmpxgrid->topterm->chrlayer);
    }
  }

  if ( subnet->chyl == (channel->lowchy + 1) && subnet->net == botnet ) {
    tmpxcrd = PERMfindXcrd(tmpxgrid->botterm);
    tmpvia = channel->ch_matrix[subnet->chxl][channel->lowchy + 1].via;
    if ( tmpvia == NULL && tmpxcrd != tmpxgrid->xcoord ) {
      /* make the extra horizontal subnet */
      xlcrd = min(tmpxcrd, tmpxgrid->xcoord);
      xhcrd = max(tmpxcrd, tmpxgrid->xcoord);
      ylcrd = yhcrd = (channel->lowchy + 1) * minctctsp;
      ROUTaddSnet(channel, xlcrd, ylcrd, xhcrd, yhcrd,
		  tmpxgrid->botterm->chrlayer);
    }
  }

  if ( subnet->chyh == channel->hichy ) {
    if ( subnet->chyl != (channel->hichy - 1) || subnet->net != topnet ) {
      fprintf(stderr,"ROUTvertXcrd;1: internal error\n");
      exit(-1);
    }
    tmpxcrd = PERMfindXcrd(tmpxgrid->topterm);
    tmpvia = channel->ch_matrix[subnet->chxl][channel->hichy - 1].via;
    ylcrd = yhcrd = (channel->hichy - 1) * minctctsp;
  }
  else if ( subnet->chyl == channel->lowchy ) {
    if ( subnet->chyh != (channel->lowchy + 1) || subnet->net != botnet ) {
      fprintf(stderr,"ROUTvertXcrd;2: internal error\n");
      /*
	exit(-1);
	*/
    }
    tmpxcrd = PERMfindXcrd(tmpxgrid->botterm);
    tmpvia = channel->ch_matrix[subnet->chxh][channel->lowchy + 1].via;
    ylcrd = yhcrd = (channel->lowchy + 1) * minctctsp;
  }
  else {
    return ( tmpxgrid->xcoord );
  }

  if ( tmpvia != NULL && tmpxcrd != tmpxgrid->xcoord ) {
    /* make the extra horizontal subnet */
    xlcrd = min(tmpxcrd, tmpxgrid->xcoord);
    xhcrd = max(tmpxcrd, tmpxgrid->xcoord);
    ROUTaddSnet(channel, xlcrd, ylcrd, xhcrd, yhcrd, subnet->layer);
  }

  return( tmpxcrd );
}

ROUTaddSnet(channel, xlcrd, ylcrd, xhcrd, yhcrd, layer)
struct _channel *channel;
int xlcrd, ylcrd, xhcrd, yhcrd;
int layer;
{
  int lowleftx, lowlefty, uprightx, uprighty;
  int width;
  char *layername;


  if ( layer == 1 ) {
    width = polywidth;
    layername = botpoly_L;
  }
  else if ( layer == 2 ) {
    width = metwidth;
    layername = metal_L;
  }
  else {
    fprintf(stderr,"ROUTgenLay;1: internal error\n");
    exit(-1);
  }

  lowleftx = xlcrd - width/2;
  lowlefty = ylcrd - width/2;
  uprightx = xhcrd + width - width/2;
  uprighty = yhcrd + width - width/2;
  makebox(&Ochannel,layername,lowleftx,lowlefty, uprightx,uprighty);
}


ROUThorizXcrd(channel, subnet, endflag)
struct _channel *channel;
struct _subnet *subnet;
int endflag; /* LEFT or RIGHT */
{
  int topnet, botnet;
  int chx, chy;
  int termxcrd;

  struct _xgrids *tmpxgrid;
  struct _terminal *tmpterm;


  if ( endflag == LEFT ) {
    chx = subnet->chxl;
    chy = subnet->chyl;
  }
  else if ( endflag == RIGHT ) {
    chx = subnet->chxh;
    chy = subnet->chyh;
  }
  else {
    return( -1 );
  }

  tmpxgrid = channel->xgrids + chx;
  botnet = topnet = -1;
  if ( tmpxgrid->botterm != NULL ) {
    botnet = tmpxgrid->botterm->netptr->number;
  }
  if ( tmpxgrid->topterm != NULL ) {
    topnet = tmpxgrid->topterm->netptr->number;
  }

  if ( chy == (channel->lowchy + 1) && subnet->net == botnet ) {
    tmpterm = tmpxgrid->botterm;
    termxcrd = PERMfindXcrd(tmpxgrid->botterm);
  }
  else if ( chy == (channel->hichy - 1) && subnet->net == topnet ) {
    tmpterm = tmpxgrid->topterm;
    termxcrd = PERMfindXcrd(tmpxgrid->topterm);
  }
  else {
    return( tmpxgrid->xcoord );
  }

  if ( tmpterm->chrlayer == subnet->layer ) {
    if ( endflag == LEFT &&
	termxcrd < tmpxgrid->xcoord &&
	channel->ch_matrix[chx][chy].via != NULL ) {

      return( tmpxgrid->xcoord );

    }
    if ( endflag == RIGHT &&
	termxcrd > tmpxgrid->xcoord &&
	channel->ch_matrix[chx][chy].via != NULL ) {

      return( tmpxgrid->xcoord );

    }

    return ( termxcrd );

  }
  else {

    return( tmpxgrid->xcoord );

  }
}
