/* the following program generates the channel router information for shields */
/* Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli */
/* Last Change: 8/11/87 */

/*
 * Modifications to Adore
 * Henry Chang (henryc@ic.Berkeley.EDU)
 * Fall, 1990
 */

#define C_FILE_NAME "ROUTshieldIn.c"

#include <stdio.h>
#include "ADglobals.h"
#include "TEKtech.h"
#include "ROUTglobals.h"

/* generate the input file for shield routing */
ROUTshieldIn(channel,tmp_dir)
struct _channel *channel;
char *tmp_dir;
{
	FILE *fp, *fopen();
	int counter;
	int num_isolated; /* number of isolated grounded pins */
	char filename[FILENAME_SIZE];
			

	ROUTshgrTerms( channel );
	ROUTshCrossOvers( channel );
	ROUTshieldNets( channel );

	num_isolated = 0;
	for ( counter = 1; counter <= channel->numshgr; counter++) {
		if ( channel->isolated_grs[counter] == YES ) {
			num_isolated++;
		}
	}

	sprintf (filename, "%s/%s", tmp_dir, SH_IN_FILE);
	fp = fopen(filename,"w");

	fprintf(fp,"channeldef\n");

	fprintf(fp,"rectagoncorners %d\n", 4);
	fprintf(fp,"%d %d\n", channel->lowchx, channel->lowchy);
	fprintf(fp,"%d %d\n", channel->hichx, channel->lowchy); 
	fprintf(fp,"%d %d\n", channel->hichx, channel->hichy);
	fprintf(fp,"%d %d\n", channel->lowchx, channel->hichy);

	fprintf(fp,"number_of_nets %d\n", channel->numshgr - num_isolated);
	fprintf(fp,"number_of_pins %d\n", channel->numshpins - num_isolated);

	for ( counter = 1; counter <= channel->numshpins; counter++) {
		if (counter<=channel->numshgr && channel->isolated_grs[counter] == YES){
			/* don't need to add the pin to the list */
		}
		else {
			fprintf(fp,"%d %d %d %d\n",channel->shpins[counter].shieldnet,
						 channel->shpins[counter].xcoord,
							 channel->shpins[counter].ycoord, 2);
		}
	}

	fprintf(fp,"obstacles %d\n", 2 * channel->numvias);

	for ( counter = 1; counter <= channel->numvias; counter++) {
		fprintf(fp,"%d %d %d %d %d\n",channel->vialist[counter].chx,
										channel->vialist[counter].chy,
										channel->vialist[counter].chx,
										channel->vialist[counter].chy, 2);
		fprintf(fp,"%d %d %d %d %d\n",channel->vialist[counter].chx,
										channel->vialist[counter].chy,
										channel->vialist[counter].chx, 
										channel->vialist[counter].chy, 1);
	}

	fprintf(fp,".end\n");
	fclose( fp );
}


/* assign net numbers to shield terminals */
ROUTshieldNets(channel)
struct _channel *channel;
{
	int counter;
	int curgr; /* the net number of the current grounded sheild pin */
	int curdist, nxtdist; /* x-distance to current and next gounded shield */
	/* used for bubble sort */
	int pass, intchnge, curorder, nextorder, tempcoord;


	/*** the flag for a given pin is YES if it is an isolated ground pin */
	/*** each ground is identifed by its net number which is between 1
	and number of ground pins */

	CALLOC(channel->isolated_grs, channel->numshgr + 1, int);

	/*** need to sort the grounded shields according to their x coordinates */
	pass = 1;
	intchnge = YES;
	while ( pass <= channel->numshgr - 1 && intchnge == YES ) {
		intchnge = NO;
		for ( counter = 1; counter<= channel->numshgr - pass; counter++) {
			curorder = channel->shpins[counter].xcoord;
			nextorder = channel->shpins[counter + 1].xcoord;
			if ( curorder > nextorder ) {
				intchnge = YES;
				tempcoord = channel->shpins[counter + 1].xcoord;
				channel->shpins[counter+1].xcoord = 
							channel->shpins[counter].xcoord;
				channel->shpins[counter].xcoord = tempcoord;

				tempcoord = channel->shpins[counter + 1].ycoord;
				channel->shpins[counter+1].ycoord = 
							channel->shpins[counter].ycoord;
				channel->shpins[counter].ycoord = tempcoord;
			}
		}
		pass++;
	}

	/*** assign pins to grounded pins, the assignment is made to the closest
	grounded pin */
	/*** the grounded shileds are increasing order, so are the remaining pins */
	curgr = 1;
	for (counter = 1; counter <= channel->numshpins; counter++) {
		if ( counter <= channel->numshgr ) {
			/* assign the net numbers to the grounde pins */
			channel->shpins[counter].shieldnet = counter;
			channel->isolated_grs[counter] = YES;
		}
		else {
			while ( curgr < channel->numshgr ) {
				curdist = abs(channel->shpins[counter].xcoord - 
											 channel->shpins[curgr].xcoord);
				nxtdist = abs(channel->shpins[counter].xcoord - 
											 channel->shpins[curgr+1].xcoord);
				if ( curdist <= nxtdist ) {
					break;
				}
				else {
					curgr++;
				}
			}
			channel->shpins[counter].shieldnet = curgr;
			channel->isolated_grs[curgr] = NO;
		}
	}
}


/* find the ground connections for each channel */
ROUTshgrTerms( channel )
struct _channel *channel;
{
  int row, column;
  int minsp; /* minimum spacing between center of contact and adjacent poly */


/*** the minimum spacing between the contact and the poly line */
  minsp = polyconsp + (polywidth + mincontd) /2;

  CALLOC(channel->shpins, channel->numwires + 100, struct _shterm); /* !!! */
  channel->numshpins = 0;


  if ( channel->name == TOP ) {
/* generate the ground shield connections for the top channel */
    row = channel->hichy;
    for ( column = channel->lowchx + 1; column <= channel->hichx - 1;
	 column++ ) {
      if ( channel->xgrids[column].topterm == NULL &&
	  channel->ch_matrix[column][row - 1].via == NULL ) {
	
	if ( (channel->xgrids[column - 1].topterm != NULL &&
	      abs(PERMfindXcrd(channel->xgrids[column-1].topterm) - 
		  channel->xgrids[column].xcoord) < minsp) ||
	    (channel->xgrids[column + 1].topterm != NULL &&
	     abs(PERMfindXcrd(channel->xgrids[column+1].topterm) - 
		 channel->xgrids[column].xcoord) < minsp) ) {
/* the grid is too close to an adjacent pin connection,
   do not add to the list */
		
	}

	else {
	  (channel->numshpins)++;
	  channel->shpins[channel->numshpins].xcoord = column;
	  channel->shpins[channel->numshpins].ycoord = row;
	}
      }
    }
  }

  else if ( channel->name == BOTTOM ) {
    row = channel->lowchy;
    for ( column = channel->lowchx + 1; column <= channel->hichx - 1;
	 column++ ) {
      if ( channel->xgrids[column].botterm != NULL &&
	  channel->xgrids[column].botterm->netptr->sensitive == YES &&
	  channel->ch_matrix[column][row + 1].via == NULL ) {
	
	(channel->numshpins)++;
	channel->shpins[channel->numshpins].xcoord = column;
	channel->shpins[channel->numshpins].ycoord = row;
      }
    }
  }

  else {
    fprintf( stderr,"ROUTshgrTerms;1: internal error\n");
    fprintf(stderr,"continuing\n");
  }

  channel->numshgr = channel->numshpins;
}


/* detect the crossovers */
ROUTshCrossOvers( channel )
struct _channel *channel;
{
	int row, column;
	int width;
	int lowleftx, lowlefty, uprightx, uprighty;
	int sensitive, insensitive;
	struct _snet_stack *tmpsnet_stack;
	struct _subnet *tmpsnet;


	width = max(metwidth, polywidth);

	for ( column = channel->lowchx; column<= channel->hichx; column++) {
		for ( row = channel->lowchy + 1; row<= channel->hichy - 1; row++ ) {
			sensitive = insensitive = NO;
			tmpsnet_stack = channel->ch_matrix[column][row].snet_stack;
			for (; tmpsnet_stack != NULL; tmpsnet_stack = tmpsnet_stack->next) {
				if (ordered_nets[tmpsnet_stack->subnet->net]->sensitive == YES){
					sensitive = YES;
				}
				else {
					insensitive = YES;
				}
			}
					
			if ( sensitive == YES && insensitive == YES) {
				if ( channel->ch_matrix[column][row].via != NULL ) {
					fprintf(stderr,"ROUTshieldTerms;2: internal error\n");
					fprintf(stderr,"continuing\n");
				}
				/* sensitive and insensitive nets cross */
				(channel->numshpins)++;
				channel->shpins[channel->numshpins].xcoord = column;
				channel->shpins[channel->numshpins].ycoord = row;

				/* here I am using botextop as topexbot */
				/*
				lowleftx = channel->xgrids[column].xcoord - width/2 - botextop;
				lowlefty = row * minctctsp - width/2 - botextop;
				uprightx = lowleftx + width + 2 * botextop;
				uprighty = lowlefty + width + 2 * botextop;
				makebox(&Ochannel,toppoly_L,lowleftx, lowlefty,
													  uprightx,uprighty);
				*/
			}
		}
	}
}
