/*
 *   set.c -- routines for maniuplating sets and set families
 */

/* LINTLIBRARY */

#include <stdio.h>
#include "UTILset.h"

extern char *malloc();

static int bit_count[256] = {
  0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
  1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
  1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
  2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
  1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
  2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
  2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
  3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8
};

#define SET_NWORDS(size)	\
    ((size) <= BPI ? 1 : (WHICH_WORD((size)-1) + 1))

#define MALLOC_SIZE(nwords)	\
    (unsigned) (sizeof(set_t) + sizeof(unsigned) * (nwords-1))


#if BPI == 32
#define count_ones(v)		\
    (bit_count[v & 255] + bit_count[(v >> 8) & 255]\
    + bit_count[(v >> 16) & 255] + bit_count[(v >> 24) & 255])
#else
#define count_ones(v)   	\
    (bit_count[v & 255] + bit_count[(v >> 8) & 255])
#endif

/* set_new -- allocate (and clear) a new set */
set_t *set_new(size)
register int size;
{
    set_t *s;
    int nwords;

    nwords = SET_NWORDS(size);
    s = (set_t *) malloc(MALLOC_SIZE(nwords));
    s->size = size;
    s->nwords = nwords;
    SET_CLEAR(s);
    return s;
}

/* set_new_copy -- allocate a new set as a copy of an existing set */
set_t *set_new_copy(a)
register set_t *a;
{
    set_t *s;
    int nwords = a->nwords;

    s = (set_t *) malloc(MALLOC_SIZE(nwords));
    s->size = a->size;
    s->nwords = nwords;
    SET_COPY(s, a);
    return s;
}


/* set_clear -- make a set the empty set */
void set_clear(r)
register set_t *r;
{
    SET_CLEAR(r);
}


/* set_fill -- make "r" the universal set of "size" elements */
void set_fill(r)
register set_t *r;
{
    SET_FILL(r);
}


/* set_copy -- copy set a into set r */
void set_copy(r, a)
register set_t *r, *a;
{
    SET_COPY(r, a);
}


/* set_and -- compute intersection of sets a and b */
void set_and(r, a, b)
register set_t *r, *a, *b;
{
    SET_AND(r, a, b);
}


/* set_or -- compute union of sets a and b */
void set_or(r, a, b)
register set_t *r, *a, *b;
{
    SET_OR(r, a, b);
}


/* set_diff -- compute difference between sets a and b */
void set_diff(r, a, b)
register set_t *r, *a, *b;
{
    SET_DIFF(r, a, b);
}

/* set_compare -- compare two sets (lexical order) */
int set_compare(a, b)
register set_t *a, *b;
{
    register int x;
    SET_COMPARE(a, b, x);
    return x;
}


/* setp_empty -- check if the set a is empty */
int setp_empty(a)
register set_t *a;
{
    register int x;
    SETP_EMPTY(a, x);
    return x;
}


/* setp_full -- check if the set a is the full set */
int setp_full(a)
register set_t *a;
{
    register int i;
    for(i = a->nwords-2; i >= 0; i--) {
	if (a->data[i] != (unsigned) ~ 0) {
	    return 0;
	}
    }
    if (a->data[a->nwords-1] != ODD_WORD(a)) {
	return 0;
    }
    return 1;
}


/* setp_equal -- check if the set a equals set b */
int setp_equal(a, b)
register set_t *a, *b;
{
    register int x;
    SETP_EQUAL(a, b, x);
    return x;
}


/* setp_disjoint -- check if intersection of a and b is empty */
int setp_disjoint(a, b)
register set_t *a, *b;
{
    register int x;
    SETP_DISJOINT(a, b, x);
    return x;
}


/* setp_implies -- check if a implies b (b contains a) */
int setp_implies(a, b)
register set_t *a, *b;
{
    register int x;
    SETP_IMPLIES(a, b, x);
    return x;
}


/* set_ord -- count number of elements in a set */
int set_ord(a)
register set_t *a;
{
    register int i, sum;
    register unsigned val;
    sum = 0;
    for(i = a->nwords-1; i >= 0; i--) {
        if ((val = a->data[i]) != 0) {
            sum += count_ones(val);
	}
    }
    return sum;
}

#define MAX_LEN 256
static char s1[MAX_LEN];

/* set_format -- format a set as [val1,val2,val3] */
char *set_format(a)
register set_t *a;
{
    register int i, num, l, len = 0;
    char temp[20];
    int first = 1;

    s1[len++] = '[';
    for(i = 0; i < a->size; i++) {
        if (is_in_set(a, i)) {
            if (! first) {
                s1[len++] = ',';
	    }
            first = 0; 

            /* Generate digits (reverse order) */
	    num = i;
            l = 0; do temp[l++] = num % 10 + '0'; while ((num /= 10) > 0);

            /* Copy them back in correct order */
            do s1[len++] = temp[--l]; while (l > 0);

            if (len > MAX_LEN-6) {
                s1[len++] = '.'; s1[len++] = '.'; s1[len++] = '.';
                break;
            }
        }
    }

    s1[len++] = ']';
    s1[len++] = '\0';
    return s1;
}


/* set_format_bv -- format a set as a bit-vector */
char *set_format_bv(s)
set_t *s;
{
    register int i;
    for(i = 0; i < s->size; i++) {
	s1[i] = is_in_set(s, i) ? '1' : '0';
    }
    s1[s->size] = '\0';
    return s1;
}

int set_error(error)
set_error_t error;
{
    (void) fprintf(stderr, "\nSet error:");
    switch(error) {
    case SET_SIZE_MISMATCH: 
	(void) fprintf(stderr, " set sizes do not match\n");
	break;
    case SET_NWORDS_MISMATCH:
	(void) fprintf(stderr, " set nwords do not match\n");
	break;
    case SET_NWORDS_BAD:
	(void) fprintf(stderr, " set size doesn't match nwords\n");
	break;
    case SET_DIRTY:
	(void) fprintf(stderr, " set is dirty\n");
	break;
    case SET_SUBSCRIPT:
	(void) fprintf(stderr, " set subscript out of range\n");
	break;
    }
    (void) fflush(stdout);
    abort();
}


void set_check1(a)
register set_t *a;
{
    if (a->nwords != SET_NWORDS(a->size)) set_error(SET_NWORDS_BAD);
    if (a->data[a->nwords-1] & ~ ODD_WORD(a)) set_error(SET_DIRTY);
}


void set_check2(a, b)
register set_t *a, *b;
{
    if (a->size != b->size) set_error(SET_SIZE_MISMATCH);
    if (a->nwords != b->nwords) set_error(SET_NWORDS_MISMATCH);
    if (a->nwords != SET_NWORDS(a->size)) set_error(SET_NWORDS_BAD);
#if 0
    if (a->data[a->nwords-1] & ~ ODD_WORD(a)) set_error(SET_DIRTY);
    if (b->data[b->nwords-1] & ~ ODD_WORD(b)) set_error(SET_DIRTY);
#endif
}


void set_check3(a, b, c)
register set_t *a, *b, *c;
{
    if (a->size != b->size) set_error(SET_SIZE_MISMATCH);
    if (a->size != c->size) set_error(SET_SIZE_MISMATCH);
    if (a->nwords != b->nwords) set_error(SET_NWORDS_MISMATCH);
    if (a->nwords != c->nwords) set_error(SET_NWORDS_MISMATCH);
    if (a->nwords != SET_NWORDS(a->size)) set_error(SET_NWORDS_BAD);
#if 0
    if (a->data[a->nwords-1] & ~ ODD_WORD(a)) set_error(SET_DIRTY);
    if (b->data[b->nwords-1] & ~ ODD_WORD(b)) set_error(SET_DIRTY);
    if (c->data[c->nwords-1] & ~ ODD_WORD(c)) set_error(SET_DIRTY);
#endif
}
